from enum import Enum
from typing import Set

from ...types import ExternalAsset, classproperty


class TableauAsset(ExternalAsset):
    """
    Tableau assets
    """

    CUSTOM_SQL_TABLE = "custom_sql_tables"
    CUSTOM_SQL_QUERY = "custom_sql_queries"
    DASHBOARD = "dashboards"
    DASHBOARD_SHEET = "dashboards_sheets"
    DATASOURCE = "datasources"
    FIELD = "fields"
    PROJECT = "projects"
    PUBLISHED_DATASOURCE = "published_datasources"
    SHEET = "sheets"
    USAGE = "views"
    USER = "users"
    WORKBOOK = "workbooks"
    WORKBOOK_TO_DATASOURCE = "workbooks_to_datasource"

    @classproperty
    def optional(cls) -> Set["TableauAsset"]:
        return {
            TableauAsset.DASHBOARD,
            TableauAsset.DASHBOARD_SHEET,
            TableauAsset.FIELD,
            TableauAsset.SHEET,
            TableauAsset.PUBLISHED_DATASOURCE,
        }


class TableauGraphqlAsset(Enum):
    """
    Assets which can be fetched from Tableau
    """

    BIN_FIELD = "binFields"
    CALCULATED_FIELD = "calculatedFields"
    COLUMN_FIELD = "columnFields"
    CUSTOM_SQL = "customSQLTables"
    DASHBOARD = "dashboards"
    DATASOURCE = "datasources"
    GROUP_FIELD = "groupFields"
    SHEETS = "sheets"
    WORKBOOK_TO_DATASOURCE = "workbooks"



from dataclasses import field
from typing import List, Optional

from pydantic.dataclasses import dataclass

from .constants import Urls


@dataclass
class Credentials:
    """Class to handle PowerBI rest API permissions"""

    client_id: str
    tenant_id: str
    secret: str = field(metadata={"sensitive": True})
    scopes: Optional[List[str]] = None

    def __post_init__(self):
        if self.scopes is None:
            self.scopes = [Urls.DEFAULT_SCOPE]



import logging

from ...utils import LocalStorage, from_env, write_summary
from ..abstract import (
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    FUNCTIONS_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    SQLExtractionProcessor,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
    extractable_asset_groups,
)
from .client import SnowflakeClient
from .query import SnowflakeQueryBuilder

logger = logging.getLogger(__name__)

SNOWFLAKE_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.FUNCTION: FUNCTIONS_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
    WarehouseAssetGroup.ROLE: (
        WarehouseAsset.GRANT_TO_ROLE,
        WarehouseAsset.GRANT_TO_USER,
        WarehouseAsset.ROLE,
        WarehouseAsset.USER,
    ),
    WarehouseAssetGroup.SNOWFLAKE_LINEAGE: (WarehouseAsset.COLUMN_LINEAGE,),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}

SNOWFLAKE_ACCOUNT = "CASTOR_SNOWFLAKE_ACCOUNT"
SNOWFLAKE_USER = "CASTOR_SNOWFLAKE_USER"
SNOWFLAKE_PASSWORD = "CASTOR_SNOWFLAKE_PASSWORD"  # noqa: S105
SNOWFLAKE_PRIVATE_KEY = "CASTOR_SNOWFLAKE_PRIVATE_KEY"


def _credentials(params: dict) -> dict:
    """extract Snowflake credentials"""
    password = params.get("password") or from_env(SNOWFLAKE_PASSWORD, True)
    private_key = params.get("private_key") or from_env(
        SNOWFLAKE_PRIVATE_KEY, True
    )
    common = {
        "account": params.get("account") or from_env(SNOWFLAKE_ACCOUNT),
        "user": params.get("user") or from_env(SNOWFLAKE_USER),
    }
    if password:
        return {**common, "password": password}
    if private_key:
        return {**common, "private_key": private_key}

    raise ValueError("missing password or private key")


def extract_all(**kwargs) -> None:
    """
    Extract all assets from Snowflake and store the results in CSV files
    """
    output_directory, skip_existing = common_args(kwargs)
    client = SnowflakeClient(
        credentials=_credentials(kwargs),
        warehouse=kwargs.get("warehouse"),
        role=kwargs.get("role"),
    )

    query_builder = SnowflakeQueryBuilder(
        db_allowed=kwargs.get("db_allowed"),
        db_blocked=kwargs.get("db_blocked"),
        fetch_transient=kwargs.get("fetch_transient"),
    )

    storage = LocalStorage(directory=output_directory)

    extractor = SQLExtractionProcessor(
        client=client,
        query_builder=query_builder,
        storage=storage,
    )

    for group in extractable_asset_groups(SNOWFLAKE_ASSETS):
        for asset in group:
            logger.info(f"Extracting `{asset.value.upper()}` ...")
            location = extractor.extract(asset, skip_existing)
            logger.info(f"Results stored to {location}\n")

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



from typing import Dict, List


def build_path(
    row: Dict,
    keys: List[str],
) -> str:
    """
    format an asset's path:
    - picks the given keys from dict
    - join keys with a dot "."
    """
    key_values = [row[key] for key in keys]
    return ".".join(key_values)


def tag_label(row: Dict) -> str:
    """
    format the tag's label:
    - {key:value} when the value is not empty
    - {key} otherwise
    """
    tag_name = row["tag_name"]
    tag_value = row["tag_value"]
    if not tag_value:
        return tag_name
    return f"{tag_name}:{tag_value}"



import json
from datetime import datetime, timedelta
from typing import Optional, Tuple, Union
from unittest.mock import MagicMock, patch

import pytest
from dateutil.tz import tzutc

from .client import ContentType, DbtClient, DbtRun, _account_url  # type: ignore
from .credentials import DbtCredentials

_DBT_CLIENT_PATH = (
    "source.packages.extractor.castor_extractor.utils.dbt.client.DbtClient"
)
_OLD_DATE = datetime(2023, 7, 10, 12, 6, 23, 109171, tzinfo=tzutc())
_OLD_DATE_STR = "2023-07-10 12:06:23.109171+00:00"
_RECENT_DATE = datetime(2023, 10, 6, 5, 9, 31, 731991, tzinfo=tzutc())
_RECENT_DATE_STR = "2023-10-06 05:09:31.731991+00:00"


def _assert_called_with(
    mocked_call: MagicMock,
    job_id: Union[int, str],
    date_range: Optional[Tuple[datetime, datetime]] = None,
) -> None:
    url = "https://cloud.getdbt.com/api/v2/accounts/40/runs/"
    params = {
        "job_definition_id": job_id,
        "limit": 1,
        "order_by": "-finished_at",
        "status": 10,
    }
    if date_range:
        params["finished_at__range"] = json.dumps(
            [d.isoformat() for d in date_range]
        )
    mocked_call.assert_called_with(url, params)


def test_DbtClient_last_run():
    default_job_id = "123"
    now = datetime.now()
    two_days_ago = now - timedelta(days=2)

    infer_path = f"{_DBT_CLIENT_PATH}._infer_account_id"
    call_path = f"{_DBT_CLIENT_PATH}._call"

    mock_response_default_job = [{"id": 1, "finished_at": _OLD_DATE_STR}]
    mock_response_job_42 = [{"id": 2, "finished_at": _RECENT_DATE_STR}]

    with patch(infer_path, return_value=40), patch(call_path) as mocked_call:
        credentials = DbtCredentials(token="some-token", job_id=default_job_id)

        dbt_client = DbtClient(credentials=credentials)

        # default job ID, no range
        mocked_call.return_value = mock_response_default_job
        result = dbt_client.last_run()
        assert result == DbtRun(id=1, finished_at=_OLD_DATE)
        _assert_called_with(mocked_call, default_job_id)

        # given job ID, no range
        mocked_call.return_value = mock_response_job_42
        result = dbt_client.last_run(job_id=42)
        assert result == DbtRun(id=2, finished_at=_RECENT_DATE)
        _assert_called_with(mocked_call, 42)

        # with range, empty response
        mocked_call.return_value = None
        result = dbt_client.last_run(finished_at_range=(two_days_ago, now))
        assert result is None
        _assert_called_with(mocked_call, default_job_id, (two_days_ago, now))

        # results array returned, but dict is None (shouldn't happen but who knows)
        mocked_call.return_value = [None]
        with pytest.raises(ValueError):
            dbt_client.last_run()


def test_DbtClient_list_job_identifiers():
    infer_path = f"{_DBT_CLIENT_PATH}._infer_account_id"
    call_path = f"{_DBT_CLIENT_PATH}._call"

    jobs = [
        {"id": 23, "state": 1},
        {"id": 46, "state": 2},  # deleted
        {"id": 395, "state": 1},
    ]

    with patch(infer_path, return_value=40), patch(call_path) as mocked_call:
        mocked_call.return_value = jobs
        credentials = DbtCredentials(token="some-token", job_id="1")
        dbt_client = DbtClient(credentials=credentials)

        jobs_ids = dbt_client.list_job_identifiers()
        assert jobs_ids == {23, 395}


def test_DbtClient_fetch_artifacts():
    infer_path = f"{_DBT_CLIENT_PATH}._infer_account_id"
    call_path = f"{_DBT_CLIENT_PATH}._call"
    run_id = 12345
    url = "https://cloud.getdbt.com/api/v2/accounts/40/runs/{}/artifacts/{}"

    with patch(infer_path, return_value=40), patch(call_path) as mocked_call:
        credentials = DbtCredentials(token="some-token", job_id="1")
        dbt_client = DbtClient(credentials=credentials)

        dbt_client.fetch_run_results(run_id)
        mocked_call.assert_called_with(
            url.format(run_id, "run_results.json"),
            content_type=ContentType.TEXT,
            ignore_not_found=True,
        )

        mocked_call.reset_mock()

        dbt_client.fetch_manifest(run_id)
        mocked_call.assert_called_with(
            url.format(run_id, "manifest.json"),
            content_type=ContentType.TEXT,
            ignore_not_found=False,
        )


def test___account_url():
    base_url = "https://cloud.getdbt.com"
    assert _account_url(base_url) == "https://cloud.getdbt.com/api/v2/accounts"

    base_url = "https://emea.dbt.com/"
    assert _account_url(base_url) == "https://emea.dbt.com/api/v2/accounts"



import logging
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Iterator, List, Optional, Set, Tuple

import requests

from ....utils import at_midnight, current_date, past_date, retry, timestamp_ms
from ..assets import DomoAsset
from .credentials import DomoCredentials
from .endpoints import Endpoint, EndpointFactory
from .pagination import Pagination

RawData = Iterator[dict]

DOMO_PUBLIC_URL = "https://api.domo.com"
DEFAULT_TIMEOUT = 120
TOKEN_EXPIRATION_SECONDS = timedelta(seconds=3000)  # auth token lasts 1 hour

IGNORED_ERROR_CODES = (
    HTTPStatus.BAD_REQUEST,
    HTTPStatus.NOT_FOUND,
)
ERROR_TPL = "Request failed with status code {status_code} and reason {reason}"

_RETRY_EXCEPTIONS = [
    requests.exceptions.ConnectTimeout,
    requests.exceptions.ReadTimeout,
]
_RETRY_COUNT = 2
_RETRY_BASE_MS = 10 * 60 * 1000  # 10 minutes

logger = logging.getLogger(__name__)


def _handle_response(response: requests.Response) -> requests.Response:
    response.raise_for_status()
    return response


def _ignore_or_raise(
    error: requests.RequestException,
    ignore_error_codes: Optional[Tuple[int, ...]],
) -> dict:
    """
    Raises the error unless the response status code is in the ignored error
    codes list.
    """
    if not ignore_error_codes:
        raise error

    response = error.response
    if response is None:
        raise error

    if response.status_code in ignore_error_codes:
        logger.warning(error)
        return {}

    raise error


class DomoClient:
    """
    Connect to Domo API and fetch main assets.
    https://developer.domo.com/portal/8ba9aedad3679-ap-is#platform-oauth-apis
    """

    def __init__(self, credentials: DomoCredentials):
        self._authentication = credentials.authentication
        self._bearer_headers: Optional[dict] = None
        self._session = requests.session()
        self._token_creation_time: datetime = datetime.min
        self._endpoint_factory = EndpointFactory(credentials.base_url)
        self._private_headers = credentials.private_headers
        self._timeout = DEFAULT_TIMEOUT
        self.base_url = credentials.base_url
        self.cloud_id = credentials.cloud_id

    def _token_expired(self) -> bool:
        token_lifetime = datetime.now() - self._token_creation_time
        return token_lifetime > TOKEN_EXPIRATION_SECONDS

    def _bearer_auth(self) -> dict:
        if self._bearer_headers and not self._token_expired():
            return self._bearer_headers

        logger.info("Refreshing authentication token...")

        basic_authentication = self._authentication
        endpoint = self._endpoint_factory.authentication

        response = self._session.get(
            endpoint.url(),
            auth=basic_authentication,
            timeout=self._timeout,
        )
        result = _handle_response(response).json()

        bearer_token = result["access_token"]
        self._bearer_headers = {"authorization": f"Bearer {bearer_token}"}
        self._token_creation_time = datetime.now()

        return self._bearer_headers

    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_COUNT,
        base_ms=_RETRY_BASE_MS,
    )
    def _get(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> requests.Response:
        params = params if params else {}
        is_private = endpoint.is_private
        headers = self._private_headers if is_private else self._bearer_auth()

        response = self._session.get(
            url=endpoint.url(asset_id),
            headers=headers,
            params=params,
            timeout=self._timeout,
        )

        if response.status_code != HTTPStatus.OK:
            logger.warning(
                ERROR_TPL.format(
                    status_code=response.status_code,
                    reason=response.reason,
                )
            )
        return response

    def _get_element(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
        ignore_error_codes: Optional[Tuple[int, ...]] = None,
    ) -> dict:
        """Used when the response only contains one element"""
        try:
            return self._get(endpoint, params, asset_id).json()
        except requests.RequestException as error:
            return _ignore_or_raise(error, ignore_error_codes)

    def _get_many(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> List[dict]:
        """Used when the response contains multiple elements"""
        return self._get(endpoint, params, asset_id).json()

    def _get_paginated(self, endpoint: Endpoint) -> List[dict]:
        """Used when the response is paginated and need iterations"""
        pagination = Pagination()
        all_results: List[dict] = []

        while pagination.needs_increment:
            params = {**pagination.params, **endpoint.params}
            results = self._get_many(endpoint=endpoint, params=params)
            all_results.extend(results)
            number_of_items = len(results)
            pagination.increment_offset(number_of_items)

        return all_results

    def _datasources(self, page_id: str) -> RawData:
        endpoint = self._endpoint_factory.page_content(page_id)
        page_content = self._get_element(
            endpoint,
            ignore_error_codes=IGNORED_ERROR_CODES,
        )
        processed: set[str] = set()
        for card in page_content.get("cards", []):
            for datasource in card["datasources"]:
                id_ = datasource["dataSourceId"]
                if id_ in processed:
                    continue
                yield {
                    "id": id_,
                    "name": datasource["dataSourceName"],
                    "type": datasource["dataType"],
                }
                processed.add(id_)

    def _process_pages(
        self,
        page_tree: List[dict],
        parent_path: str = "/root",
    ) -> Iterator[dict]:
        """Recursively fetch pages while building the folder architecture"""
        if not page_tree:
            return []

        for page in page_tree:
            page_id = page.get("id")
            page_name = page.get("name")
            if not (page_id and page_name):
                logger.debug(f"This page has no id nor name: {page}")
                continue

            page_children = page.get("children", list)

            detail = self._get_element(
                self._endpoint_factory.pages,
                asset_id=page_id,
            )
            datasources = self._datasources(page_id)
            yield {
                **detail,
                "datasources": list(datasources),
                "path": parent_path,
                "base_url": self.base_url,
            }
            yield from self._process_pages(
                page_tree=page_children,
                parent_path=f"{parent_path}/{page_name}",
            )

    def _link_table_to_dataset(self, dataset: dict, dataset_id: str) -> dict:
        if not self.cloud_id:
            return dataset

        lineage_endpoint = self._endpoint_factory.table_lineage(
            dataset_id=dataset_id,
            cloud_id=self.cloud_id,
        )
        dataset_table_lineage = self._get_element(lineage_endpoint)

        return {**dataset, **dataset_table_lineage}

    def _pages(self) -> RawData:
        page_hierarchy = self._get_paginated(self._endpoint_factory.pages)
        yield from self._process_pages(page_hierarchy)

    def _datasets(self) -> RawData:
        dataset_list = self._get_paginated(self._endpoint_factory.datasets)

        for dataset in dataset_list:
            dataset_id = dataset.get("id")
            if not dataset_id:
                continue

            dataset_detail = self._get_element(
                endpoint=self._endpoint_factory.datasets,
                asset_id=dataset_id,
            )

            yield {
                **self._link_table_to_dataset(
                    dataset=dataset_detail,
                    dataset_id=dataset_id,
                ),
                "base_url": self.base_url,
            }

    def _unique_datasets(self) -> RawData:
        ids_encountered: Set[str] = set()
        for dataset in self._datasets():
            dataset_id = dataset.get("id")
            if not dataset_id or dataset_id in ids_encountered:
                continue
            ids_encountered.add(dataset_id)
            yield dataset

    def _users(self) -> RawData:
        user_list = self._get_paginated(self._endpoint_factory.users)
        for user in user_list:
            user_id = user.get("id")
            if not user_id:
                continue
            yield self._get_element(
                endpoint=self._endpoint_factory.users,
                asset_id=user_id,
            )

    def _audit(self) -> RawData:
        yesterday = timestamp_ms(at_midnight(past_date(1)))
        today = timestamp_ms(at_midnight(current_date()))
        yield from self._get_paginated(
            self._endpoint_factory.audit(yesterday, today)
        )

    def _dataflows(self) -> RawData:
        dataflows = self._get_many(self._endpoint_factory.dataflows)
        for dataflow in dataflows:
            dataflow_id = dataflow.get("id")
            if not dataflow_id:
                continue
            yield self._get_element(self._endpoint_factory.lineage(dataflow_id))

    def fetch(self, asset: DomoAsset) -> RawData:
        """Returns the needed metadata for the queried asset"""
        if asset == DomoAsset.PAGES:
            yield from self._pages()

        elif asset == DomoAsset.DATASETS:
            yield from self._unique_datasets()

        elif asset == DomoAsset.USERS:
            yield from self._users()

        elif asset == DomoAsset.AUDIT:
            yield from self._audit()

        elif asset == DomoAsset.DATAFLOWS:
            yield from self._dataflows()

        else:
            raise ValueError(f"This asset {asset} is unknown")



#!/usr/bin/env python3
import logging
import ntpath
from datetime import datetime
from typing import Dict, Iterable, Optional, Tuple
from uuid import UUID

import requests

from ..utils.retry import retry
from .constant import (
    INGEST_URL,
    PATH_TEMPLATES,
    RETRY_BASE_MS,
    RETRY_JITTER_MS,
    RETRY_STRATEGY,
    FileType,
)
from .env import get_blob_env
from .utils import iter_files

logger = logging.getLogger(__name__)

_EXCEPTIONS = (
    requests.exceptions.Timeout,
    requests.exceptions.ConnectTimeout,
)


def _path_and_url(
    source_id: UUID,
    file_type: FileType,
    file_path: str,
) -> Tuple[str, str]:

    now = datetime.utcnow()
    timestamp = int(now.timestamp())
    filename = ntpath.basename(file_path)
    path_template = PATH_TEMPLATES[file_type]
    path = path_template.format(
        timestamp=timestamp,
        source_id=source_id,
        filename=filename,
    )

    url = f"{INGEST_URL}/{path}"

    return path, url


def _headers(token: str) -> Dict:
    return {
        "Authorization": f"Token {token}",
        "Accept": "text/csv, application/json",
    }


def _upload(
    token: str,
    source_id: UUID,
    file_path: str,
    file_type: FileType,
) -> None:
    """
    Upload the given file to Google Cloud Storage (GCS)
    - Don't call GCS API directly
    - Call the ingestion proxy which handles authorisation and uploading
    """
    path, url = _path_and_url(source_id, file_type, file_path)
    headers = _headers(token)
    timeout, max_retries = get_blob_env()

    with open(file_path, "rb") as file_content:

        @retry(
            exceptions=_EXCEPTIONS,
            max_retries=max_retries,
            base_ms=RETRY_BASE_MS,
            jitter_ms=RETRY_JITTER_MS,
            strategy=RETRY_STRATEGY,
        )
        def _request_post():
            response = requests.post(
                url=url,
                headers=headers,
                files={"file": file_content},
                timeout=timeout,
            )
            response.raise_for_status()

        _request_post()

    logger.info(f"Uploaded {file_path} as {file_type.value} to {path}")


def upload_manifest(
    token: str,
    source_id: UUID,
    file_path: str,
) -> None:
    """
    credentials: path to file or dict
    source_id: id for the source
    file_path: path to the local manifest to upload
    """
    _upload(token, source_id, file_path, FileType.DBT)


def upload(
    token: str,
    source_id: UUID,
    file_type: FileType,
    file_path: Optional[str] = None,
    directory_path: Optional[str] = None,
) -> None:
    """
    credentials: path to file or dict
    source_id: id for the source
    file_type: type of file(s) uploaded - see FileType Enum
    file_path: path to the local visualization or warehouse file to upload
    directory_path: path to the local directory containing files to upload
    """
    files: Iterable[str]
    if directory_path:
        files = iter_files(directory_path)
    elif file_path:
        files = [file_path]
    else:
        message = "either file_path or directory_path should be defined"
        raise ValueError(message)

    for file_ in files:
        _upload(token, source_id, file_, file_type)



from typing import Callable, Dict, List
from uuid import UUID

from .pager_on_id import PagerOnId

ITEMS_WITH_IDS = [
    {"id": "00000000-0000-0000-0000-00000000000" + str(i)} for i in range(1, 10)
]


def _make_callback_with_ids(
    elements: List[Dict[str, str]],
) -> Callable[[UUID, int], List[Dict[str, str]]]:
    def _callback(max_id: UUID, per: int) -> List[Dict[str, str]]:
        """assumes the elements are sorted by id"""
        to_return: List[Dict[str, str]] = []
        for element in elements:
            if element["id"] > str(max_id):
                to_return.append(element)
            if len(to_return) >= per:
                return to_return
        return to_return

    return _callback


def test_pageronid__all():
    """relies on ITEMS_WITH_IDS being sorted by id"""
    pager = PagerOnId(_make_callback_with_ids(ITEMS_WITH_IDS))
    # When no argument provided
    assert pager.all() == ITEMS_WITH_IDS
    # When per page is less than the number of ITEMS
    assert pager.all(per_page=1) == ITEMS_WITH_IDS
    # When per page is more than the number of ITEMS
    assert pager.all(per_page=1000) == ITEMS_WITH_IDS


def test_pageronid__iterator__pagination():
    """unit test for PagerOnId#iterator() (pagination)"""
    pager = PagerOnId(_make_callback_with_ids(ITEMS_WITH_IDS))

    def nb_of_pages(per_page: int) -> int:
        return len([page for page in pager.iterator(per_page=per_page)])

    assert nb_of_pages(per_page=50) == 1
    assert nb_of_pages(per_page=1) == len(ITEMS_WITH_IDS)
    assert nb_of_pages(per_page=2) == 5
    assert nb_of_pages(per_page=4) == 3



KEYS = {
    "CASTOR_TABLEAU_USER": "Test",
    "CASTOR_TABLEAU_SERVER_URL": "http://test",
    "CASTOR_TABLEAU_PASSWORD": "Test",
    "CASTOR_TABLEAU_SITE_ID": "dad65087-b08b-4603-af4e-2887b8aafc67",
}



from abc import ABC, abstractmethod
from typing import Iterator, Optional, cast

from sqlalchemy import text
from sqlalchemy.engine import Connection, ResultProxy, create_engine

from .query import ExtractionQuery


class AbstractSourceClient(ABC):
    """interface for the client to connect to the source"""

    def __init__(self, **credentials: dict):
        """init signature can vary"""

    @abstractmethod
    def connect(self) -> Connection:
        pass

    @abstractmethod
    def close(self, dispose: Optional[bool] = False) -> bool:
        pass

    @abstractmethod
    def execute(self, query: ExtractionQuery) -> Iterator[dict]:
        pass


class SqlalchemyClient(AbstractSourceClient, ABC):
    def __init__(self, credentials: dict):
        super().__init__(**credentials)
        self._uri = self._build_uri(credentials)
        self._options = self._engine_options(credentials)
        self._engine = create_engine(self._uri, **self._options)
        self._connection: Optional[Connection] = None

    @abstractmethod
    def _build_uri(self, credentials: dict) -> str: ...

    @abstractmethod
    def _engine_options(self, credentials: dict) -> dict: ...

    @staticmethod
    def _process_result(proxy: ResultProxy) -> Iterator[dict]:
        return (dict(row) for row in proxy)

    def connect(self) -> Connection:
        if not self._connection or self._connection.closed:
            self._connection = self._engine.connect()
        return cast(Connection, self._connection)

    def execute(self, query: ExtractionQuery) -> Iterator[dict]:
        connection = self.connect()
        try:
            proxy = connection.execute(text(query.statement), query.params)
            results = self._process_result(proxy)
            return results
        finally:
            self.close()

    def close(self, dispose: Optional[bool] = False) -> bool:
        if not self._connection:
            return False
        self._connection.close()
        if dispose:
            self._engine.dispose()
        return True



from typing import Any, Callable, Dict, List, Optional

from .pager_on_token import PagerOnToken

ITEMS = list(range(1000))

OStr = Optional[str]


def _make_callback_with_token(
    elements: List[Dict[str, str]],
) -> Callable[[OStr, int], Dict[str, Any]]:
    def _callback(page_token: OStr, per_page: int) -> Dict[str, Any]:
        """callable with a token that indicates how to retrieve the next page"""
        if page_token:
            _start = int(page_token)
        else:
            _start = 0
        _end = _start + per_page
        return {"res": elements[_start:_end], "next_page_token": str(_end)}

    return _callback


def _make_callback_with_missing_token(
    elements: List[Dict[str, str]],
) -> Callable[[OStr, int], Dict[str, Any]]:
    def _callback(page_token: OStr, per_page: int) -> Dict[str, Any]:
        """callable with a token that indicates how to retrieve the next page
        except for the last page"""
        if page_token:
            _start = int(page_token)
        else:
            _start = 0
        _end = _start + per_page
        if _end == len(ITEMS):
            return {"res": elements[_start:_end]}
        return {"res": elements[_start:_end], "next_page_token": str(_end)}

    return _callback


def test_pagerontoken__all():
    """unit test for PagerOnToken#all()"""
    pager = PagerOnToken(_make_callback_with_token(ITEMS))
    # When no argument provided
    assert pager.all() == ITEMS
    # When per page is less than the number of ITEMS
    assert pager.all(per_page=1) == ITEMS
    # When per page is more than the number of ITEMS
    assert pager.all(per_page=len(ITEMS) + 20) == ITEMS

    # Same test suite, but no token is provided at the last call
    pager = PagerOnToken(_make_callback_with_missing_token(ITEMS))
    # When no argument provided
    assert pager.all() == ITEMS
    # When per page is less than the number of ITEMS
    assert pager.all(per_page=1) == ITEMS
    # When per page is more than the number of ITEMS
    assert pager.all(per_page=len(ITEMS) + 20) == ITEMS


def test_pagerontoken__iterator__pagination():
    """unit test for PagerOnToken#iterator() (pagination)"""
    pager = PagerOnToken(_make_callback_with_token(ITEMS))

    def nb_of_pages(per_page: int) -> int:
        return len([page for page in pager.iterator(per_page=per_page)])

    assert nb_of_pages(per_page=len(ITEMS)) == 1
    assert nb_of_pages(per_page=1) == len(ITEMS)
    assert nb_of_pages(per_page=2) == len(ITEMS) // 2
    assert nb_of_pages(per_page=4) == len(ITEMS) // 4



from .client import CARDS_KEY, CARDS_KEY_DEPRECATED, ApiClient


def test__dashboard_cards():
    for key in (CARDS_KEY, CARDS_KEY_DEPRECATED):
        dashboards = [
            {
                "name": "dash_1",
                key: ["card_1", "card_2"],
            },
            {
                "name": "dash_2",
                key: ["card_3"],
            },
            {
                "name": "dash_3",
            },
        ]

        cards = list(ApiClient._dashboard_cards(dashboards))
        assert cards == ["card_1", "card_2", "card_3"]



OLDER_DATE = "2024-04-18 20:20:20.0"
CLOSER_DATE = "2024-04-19 20:20:20.0"

MOCK_TABLES_FOR_TABLE_LINEAGE = [
    {
        "id": "f51ba2ca-8cc3-4de6-8f8b-730359e8f40f",
        "schema_id": "dev.silver",
        "table_name": "analytics",
    },
    {
        "id": "4e140bdc-a67c-4b68-8a07-c684657d8b44",
        "schema_id": "dev.silver",
        "table_name": "pre_analytics",
    },
    {
        "id": "7d403198-55ea-4a40-9995-6ee2f4c79dfa",
        "schema_id": "dev.bronze",
        "table_name": "analytics",
    },
]

_RAW_LINEAGE_DEV_SILVER_ANALYTICS = {
    "upstreams": [
        {  # there could be other keys: jobInfos, notebookInfos, queryInfos
            "tableInfo": {
                "name": "pre_analytics",
                "catalog_name": "dev",
                "schema_name": "silver",
                "table_type": "PERSISTED_VIEW",  # not used
                "lineage_timestamp": OLDER_DATE,
            }
        },
        {
            "tableInfo": {
                "name": "analytics",
                "catalog_name": "dev",
                "schema_name": "bronze",
                "table_type": "PERSISTED_VIEW",  # not used
                "lineage_timestamp": CLOSER_DATE,
            }
        },
    ],
    "downstreams": [],
}
_RAW_LINEAGE_DEV_SILVER_PRE_ANALYTICS = {
    "upstreams": [],
    "downstreams": [
        {
            "tableInfo": {
                "name": "analytics",
                "catalog_name": "dev",
                "schema_name": "silver",
                "table_type": "PERSISTED_VIEW",  # not used
                "lineage_timestamp": OLDER_DATE,
            }
        },
    ],
}
_RAW_LINEAGE_DEV_BRONZE_ANALYTICS = {
    "upstreams": [],
    "downstreams": [
        {
            "tableInfo": {
                "name": "analytics",
                "catalog_name": "dev",
                "schema_name": "silver",
                "table_type": "PERSISTED_VIEW",  # not used
                "lineage_timestamp": OLDER_DATE,
            }
        },
    ],
}

# should be in the same order as MOCK_TABLES_FOR_TABLE_LINEAGE
TABLE_LINEAGE_SIDE_EFFECT: tuple = (
    _RAW_LINEAGE_DEV_SILVER_ANALYTICS,
    _RAW_LINEAGE_DEV_SILVER_PRE_ANALYTICS,
    _RAW_LINEAGE_DEV_BRONZE_ANALYTICS,
)



import logging
from typing import Dict, Iterator, List, Tuple

from tqdm import tqdm  # type: ignore

from ...utils.salesforce import SalesforceBaseClient, SalesforceCredentials
from .format import SalesforceFormatter
from .soql import SOBJECT_FIELDS_QUERY_TPL, SOBJECTS_QUERY_TPL

logger = logging.getLogger(__name__)


class SalesforceClient(SalesforceBaseClient):
    """
    Salesforce API client to extract sobjects
    """

    # Implicit (hard-coded in Salesforce) limitation when using SOQL of 2,000 rows
    LIMIT_RECORDS_PER_PAGE = 2000

    def __init__(self, credentials: SalesforceCredentials):
        super().__init__(credentials)
        self.formatter = SalesforceFormatter()

    @staticmethod
    def name() -> str:
        return "Salesforce"

    def _format_query(self, query_template: str, start_durable_id: str) -> str:
        return query_template.format(
            start_durable_id=start_durable_id,
            limit=self.LIMIT_RECORDS_PER_PAGE,
        )

    def _next_records(
        self, url: str, query_template: str, start_durable_id: str = "0000"
    ) -> List[dict]:
        query = self._format_query(
            query_template, start_durable_id=start_durable_id
        )
        records, _ = self._call(
            url, params={"q": query}, processor=self._query_processor
        )
        return records

    def _is_last_page(self, records: List[dict]) -> bool:
        return len(records) < self.LIMIT_RECORDS_PER_PAGE

    def _should_query_next_page(
        self, records: List[dict], page_number: int
    ) -> bool:
        return not (
            self._is_last_page(records)
            or self._has_reached_pagination_limit(page_number)
        )

    def _query_all(self, query_template: str) -> Iterator[dict]:
        """
        Run a SOQL query over salesforce API

        Note, pagination is performed via a LIMIT in the SOQL query and requires
        that ids are sorted. The SOQL query must support `limit` and
        `start_durable_id` as parameters.
        """
        url = self.query_url
        logger.info("querying page 0")
        records = self._next_records(url, query_template)
        yield from records

        page_count = 1
        while self._should_query_next_page(records, page_count):
            logger.info(f"querying page {page_count}")
            last_durable_id = records[-1]["DurableId"]
            records = self._next_records(
                url, query_template, start_durable_id=last_durable_id
            )
            yield from records
            page_count += 1

    def fetch_sobjects(self) -> List[dict]:
        """Fetch all sobjects"""
        logger.info("Extracting sobjects")
        return list(self._query_all(SOBJECTS_QUERY_TPL))

    def fetch_fields(self, sobject_name: str) -> List[dict]:
        """Fetches fields of a given sobject"""
        query = SOBJECT_FIELDS_QUERY_TPL.format(
            entity_definition_id=sobject_name
        )
        response = self._call(self.tooling_url, params={"q": query})
        return response["records"]

    def tables(self) -> List[dict]:
        """
        Get Salesforce sobjects as tables
        """
        sobjects = self.fetch_sobjects()
        logger.info(f"Extracted {len(sobjects)} sobjects")
        return list(self.formatter.tables(sobjects))

    def columns(
        self, sobject_names: List[Tuple[str, str]], show_progress: bool = True
    ) -> List[dict]:
        """
        Get salesforce sobject fields as columns
        show_progress: optionally deactivate the tqdm progress bar
        """
        sobject_fields: Dict[str, List[dict]] = dict()
        for api_name, table_name in tqdm(
            sobject_names, disable=not show_progress
        ):
            fields = self.fetch_fields(api_name)
            sobject_fields[table_name] = fields
        return self.formatter.columns(sobject_fields)



from .column import ColumnChecker
from .enums import DataType, Issue


def test__column_checker():
    """Test the check() method of ColumnChecker"""

    # test data-type: float
    checker = ColumnChecker(data_type=DataType.FLOAT)
    assert checker.check("foo") == Issue.WRONG_DATATYPE
    assert checker.check("179.5") is None

    # test data-type: int
    checker = ColumnChecker(data_type=DataType.INTEGER)
    assert checker.check("bar") == Issue.WRONG_DATATYPE
    assert checker.check("-56") is None

    # test data-type: datetime
    checker = ColumnChecker(data_type=DataType.DATETIME)
    assert checker.check("calendar") == Issue.WRONG_DATATYPE
    assert checker.check("2021-12-01") is None
    assert checker.check("4/4/99") is None

    # test mandatory
    checker = ColumnChecker(is_mandatory=True)
    assert checker.check("foo") is None
    assert checker.check("") == Issue.MISSING_VALUE
    assert checker.check(None) == Issue.MISSING_VALUE

    # test unicity
    checker = ColumnChecker(is_unique=True)
    assert checker.check("foo") is None
    assert checker.check("bar") is None
    assert checker.check("foo") == Issue.DUPLICATE_VALUE

    # test foreign keys
    ids = {"1", "2", "3", "4", "5"}
    checker = ColumnChecker(foreign=ids)
    assert checker.check("1") is None
    assert checker.check("4") is None
    assert checker.check("99") == Issue.UNKNOWN_REFERENCE

    # test enums
    checker = ColumnChecker(enum_values={"TABLE", "COLUMN"})
    assert checker.check("COLUMN") is None
    assert checker.check("TABLE") is None
    assert checker.check("VIEW") == Issue.UNAUTHORIZED_VALUE

    # test data-type: list
    checker = ColumnChecker(data_type=DataType.LIST, is_mandatory=True)
    assert checker.check("") == Issue.MISSING_VALUE
    assert checker.check(None) == Issue.MISSING_VALUE
    assert checker.check("['foo', 'bar']") is None
    assert checker.check('["foo", "bar"]') is None



from typing import Any, Dict, Iterator, List

from .constants import SCHEMA_NAME


def _clean(raw: str) -> str:
    return raw.strip('"')


def _field_description(field: Dict[str, Any]) -> str:
    context: Dict[str, str] = {}

    field_definition: Dict[str, str] = field.get("FieldDefinition") or {}
    if description := field_definition.get("Description"):
        context["Description"] = _clean(description)
    if help_text := field.get("InlineHelpText"):
        context["Help Text"] = _clean(help_text)
    if compliance_group := field_definition.get("ComplianceGroup"):
        context["Compliance Categorization"] = _clean(compliance_group)
    if security_level := field_definition.get("SecurityClassification"):
        context["Data Sensitivity Level"] = _clean(security_level)

    return "\n".join([f"- {k}: {v}" for k, v in context.items()])


def _to_column_payload(field: dict, position: int, table_name: str) -> dict:
    field_name = field["QualifiedApiName"]
    return {
        "id": f"{table_name}.{field_name}",
        "table_id": table_name,
        "column_name": field_name,
        "description": _field_description(field),
        "data_type": field.get("DataType"),
        "ordinal_position": position,
    }


def _to_table_payload(sobject: dict, table_name: str) -> dict:
    return {
        "id": table_name,
        "api_name": sobject["QualifiedApiName"],
        "label": sobject["Label"],
        "schema_id": SCHEMA_NAME,
        "table_name": table_name,
        "description": "",
        "tags": [],
        "type": "TABLE",
    }


def _merge_label_and_api_name(sobject: dict) -> dict:
    label = sobject["Label"]
    api_name = sobject["QualifiedApiName"]
    table_name = f"{label} ({api_name})"
    return _to_table_payload(sobject, table_name)


def _by_label(sobjects: List[dict]) -> Dict[str, List[dict]]:
    by_label: Dict[str, List[dict]] = dict()
    for sobject in sobjects:
        label = sobject["Label"]
        similar_sobjects = by_label.setdefault(label, [])
        similar_sobjects.append(sobject)
    return by_label


class SalesforceFormatter:
    """
    Helper functions that format the response in the format to be exported as
    csv.
    """

    @staticmethod
    def tables(sobjects: List[dict]) -> Iterator[dict]:
        """
        formats the raw list of sobjects to tables
        if two tables share the same label, then we add the api name as well
        """
        by_label = _by_label(sobjects)
        for label, similars in by_label.items():
            if len(similars) > 1:
                yield from [_merge_label_and_api_name(s) for s in similars]
            else:
                sobject = similars[0]  # unique sobject on label
                yield _to_table_payload(sobject, label)

    @staticmethod
    def columns(sobject_fields: Dict[str, List[dict]]) -> List[dict]:
        """formats the raw list of sobject fields to columns"""
        return [
            _to_column_payload(field, idx, table_name)
            for table_name, fields in sobject_fields.items()
            for idx, field in enumerate(fields)
        ]



from enum import Enum
from typing import Dict

from .....utils import from_env


class CredentialsApiKey(Enum):
    BASE_URL = "base_url"
    USER = "user"
    PASSWORD = "password"  # noqa: S105


CREDENTIALS_ENV: Dict[CredentialsApiKey, str] = {
    CredentialsApiKey.BASE_URL: "CASTOR_METABASE_API_BASE_URL",
    CredentialsApiKey.USER: "CASTOR_METABASE_API_USERNAME",
    CredentialsApiKey.PASSWORD: "CASTOR_METABASE_API_PASSWORD",
}


def get_value(key: CredentialsApiKey, kwargs: dict) -> str:
    """
    Returns the value of the given key:
    - from kwargs in priority
    - from ENV if not provided (raises an error if not found in ENV)
    """
    env_key = CREDENTIALS_ENV[key]
    return str(kwargs.get(key.value) or from_env(env_key))


class CredentialsApi:
    """ValueObject for the credentials"""

    def __init__(self, base_url: str, user: str, password: str):
        self.base_url = base_url
        self.user = user
        self.password = password

    def to_dict(self, hide: bool = False) -> Dict[str, str]:
        safe = (CredentialsApiKey.BASE_URL, CredentialsApiKey.USER)
        unsafe = (CredentialsApiKey.PASSWORD,)

        def val(k: CredentialsApiKey, v: str) -> str:
            return "*" + v[-4:] if hide and k in unsafe else v

        return {a.value: val(a, getattr(self, a.value)) for a in safe + unsafe}



from typing import Any, Callable, Dict, Iterator, Optional, Sequence, TypeVar

from .pager import (
    DEFAULT_PER_PAGE,
    AbstractPager,
    PagerLogger,
    PagerStopStrategy,
)

T = TypeVar("T")


class PagerOnToken(AbstractPager):
    """Token-based pagination"""

    def __init__(
        self,
        callback: Callable[[Optional[str], Optional[int]], Dict[Any, Any]],
        *,
        logger: Optional[PagerLogger] = None,
        start_page: int = 1,
        stop_strategy: PagerStopStrategy = PagerStopStrategy.EMPTY_PAGE,
    ):
        self._callback = callback
        self._logger = logger or PagerLogger()
        self._start_page = start_page
        self._stop_strategy = stop_strategy

    def iterator(
        self,
        per_page: int = DEFAULT_PER_PAGE,
    ) -> Iterator[Sequence[T]]:
        """Yields data provided by the callback as a list using the nexttoken"""
        stop_on_empty_page = self._stop_strategy == PagerStopStrategy.EMPTY_PAGE
        page = self._start_page
        total_results = 0
        page_token = None
        while True:
            results = self._callback(page_token, per_page)
            nb_results = len(results.get("res", []))
            total_results += nb_results

            if results.get("res"):
                yield results["res"]

            stop = self.should_stop(nb_results, per_page, stop_on_empty_page)
            page_token = results.get("next_page_token")
            if stop or not page_token:
                break
            page += 1

        self._logger.on_success(page, total_results)



from datetime import date
from unittest.mock import Mock, patch

from freezegun import freeze_time

from ..abstract.time_filter import TimeFilter
from .client import DatabricksClient, LineageLinks, _day_hour_to_epoch_ms
from .test_constants import (
    CLOSER_DATE,
    MOCK_TABLES_FOR_TABLE_LINEAGE,
    OLDER_DATE,
    TABLE_LINEAGE_SIDE_EFFECT,
)


def test__day_hour_to_epoch_ms():
    _day_hour_to_epoch_ms(date(2023, 2, 14), 14) == 1644847200000


@freeze_time("2023-7-4")
def test_DatabricksClient__time_filter():
    # default is yesterday
    default_time_filter = None
    default_filter = DatabricksClient._time_filter(default_time_filter)
    expected_default = {
        "filter_by": {
            "query_start_time_range": {
                "end_time_ms": 1688428800000,  # July 4, 2023 12:00:00 AM GMT
                "start_time_ms": 1688342400000,  # July 3, 2023 12:00:00 AM GMT
            }
        }
    }
    assert default_filter == expected_default

    # custom time (from execution_date in DAG for example)
    time_filter = TimeFilter(day=date(2020, 10, 15))
    custom_filter = DatabricksClient._time_filter(time_filter)
    expected_custom = {
        "filter_by": {
            "query_start_time_range": {
                "end_time_ms": 1602806400000,  # October 16, 2020 12:00:00 AM
                "start_time_ms": 1602720000000,  # October 15, 2020 12:00:00 AM
            }
        }
    }
    assert custom_filter == expected_custom

    # hourly extraction: note that hour_min == hour_max
    hourly = TimeFilter(day=date(2023, 4, 14), hour_min=4, hour_max=4)
    hourly_filter = DatabricksClient._time_filter(hourly)
    expected_hourly = {
        "filter_by": {
            "query_start_time_range": {
                "end_time_ms": 1681448400000,  # April 14, 2023 5:00:00 AM
                "start_time_ms": 1681444800000,  # April 14, 2023 4:00:00 AM
            }
        }
    }
    assert hourly_filter == expected_hourly


class MockDatabricksClient(DatabricksClient):
    def __init__(self):
        self._db_allowed = ["prd", "staging"]
        self._db_blocked = ["dev"]


def test_DatabricksClient__keep_catalog():
    client = MockDatabricksClient()
    assert client._keep_catalog("prd")
    assert client._keep_catalog("staging")
    assert not client._keep_catalog("dev")
    assert not client._keep_catalog("something_unknown")


def test_DatabricksClient__get_user_mapping():
    client = MockDatabricksClient()
    users = [
        {"id": "both", "email": "hello@world.com", "user_name": "hello world"},
        {"id": "no_email", "email": "", "user_name": "no email"},
        {"id": "no_name", "email": "no@name.fr", "user_name": ""},
        {"id": "no_both", "email": "", "user_name": ""},
        {"id": "", "email": "no@id.com", "user_name": "no id"},
    ]
    expected = {
        "hello@world.com": "both",
        "hello world": "both",
        "no@name.fr": "no_name",
        "no email": "no_email",
    }
    mapping = client._get_user_mapping(users)
    assert mapping == expected


def test_DatabricksClient__match_table_with_user():
    client = MockDatabricksClient()
    user_mapping = {"bob@castordoc.com": 3}

    table = {"id": 1, "owner_email": "bob@castordoc.com"}
    table_with_owner = client._match_table_with_user(table, user_mapping)

    assert table_with_owner == {**table, "owner_external_id": 3}

    table_without_owner = {"id": 1, "owner_email": None}
    actual = client._match_table_with_user(table_without_owner, user_mapping)
    assert actual == table_without_owner


@patch(
    "source.packages.extractor.castor_extractor.warehouse.databricks.client.DatabricksClient.get",
    side_effect=TABLE_LINEAGE_SIDE_EFFECT,
)
def test_DatabricksClient_table_lineage(mock_get):
    client = DatabricksClient(Mock())

    lineage = client.table_lineage(MOCK_TABLES_FOR_TABLE_LINEAGE)
    assert len(lineage) == 2

    expected_link_1 = {
        "parent_path": "dev.silver.pre_analytics",
        "child_path": "dev.silver.analytics",
        "timestamp": OLDER_DATE,
    }
    expected_link_2 = {
        "parent_path": "dev.bronze.analytics",
        "child_path": "dev.silver.analytics",
        "timestamp": CLOSER_DATE,
    }
    assert expected_link_1 in lineage
    assert expected_link_2 in lineage


def test_LineageLinks_add():
    links = LineageLinks()
    timestamped_link = ("parent", "child", None)
    expected_key = ("parent", "child")

    links.add(timestamped_link)

    assert expected_key in links.lineage
    assert links.lineage[expected_key] is None

    # we replace None by an actual timestamp
    timestamped_link = ("parent", "child", OLDER_DATE)
    links.add(timestamped_link)
    assert expected_key in links.lineage
    assert links.lineage[expected_key] == OLDER_DATE

    # we update with the more recent timestamp
    timestamped_link = ("parent", "child", CLOSER_DATE)
    links.add(timestamped_link)
    assert expected_key in links.lineage
    assert links.lineage[expected_key] == CLOSER_DATE

    # we keep the more recent timestamp
    timestamped_link = ("parent", "child", OLDER_DATE)
    links.add(timestamped_link)
    assert expected_key in links.lineage
    assert links.lineage[expected_key] == CLOSER_DATE



from typing import Dict

from ...utils import from_env
from .constants import Keys

_USERNAME = "CASTOR_SALESFORCE_USERNAME"
_PASSWORD = "CASTOR_SALESFORCE_PASSWORD"  # noqa: S105
_SECURITY_TOKEN = "CASTOR_SALESFORCE_SECURITY_TOKEN"  # noqa: S105
_CLIENT_ID = "CASTOR_SALESFORCE_CLIENT_ID"
_CLIENT_SECRET = "CASTOR_SALESFORCE_CLIENT_SECRET"  # noqa: S105
_BASE_URL = "CASTOR_SALESFORCE_BASE_URL"


class SalesforceCredentials:
    """
    Class to handle Salesforce rest API permissions
    """

    def __init__(
        self,
        *,
        username: str,
        password: str,
        security_token: str,
        client_id: str,
        client_secret: str,
        base_url: str,
    ):
        self.username = username
        self.password = password + security_token
        self.client_id = client_id
        self.client_secret = client_secret
        self.base_url = base_url

    def token_request_payload(self) -> Dict[str, str]:
        """
        Params to post to the API in order to retrieve the authentication token
        """
        return {
            "grant_type": "password",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": self.username,
            "password": self.password,
        }


def to_credentials(params: dict) -> SalesforceCredentials:
    """extract Salesforce credentials"""
    username = params.get(Keys.USERNAME) or from_env(_USERNAME)
    password = params.get(Keys.PASSWORD) or from_env(_PASSWORD)
    security_token = params.get(Keys.SECURITY_TOKEN) or from_env(
        _SECURITY_TOKEN
    )
    client_id = params.get(Keys.CLIENT_ID) or from_env(_CLIENT_ID)
    client_secret = params.get(Keys.CLIENT_SECRET) or from_env(_CLIENT_SECRET)
    base_url = params.get(Keys.BASE_URL) or from_env(_BASE_URL)
    return SalesforceCredentials(
        username=username,
        password=password,
        client_id=client_id,
        client_secret=client_secret,
        security_token=security_token,
        base_url=base_url,
    )



from .api import APIClient


def test_APIClient_build_url():
    expected = "https://3.14.azuredatabricks.net/api/2.1/unity-catalog/tables"

    path = "api/2.1/unity-catalog/tables"

    host = "3.14.azuredatabricks.net"
    assert expected == APIClient.build_url(host, path)

    host_with_http = "https://3.14.azuredatabricks.net"
    assert expected == APIClient.build_url(host_with_http, path)

    host_with_trailing_slash = "https://3.14.azuredatabricks.net/"
    assert expected == APIClient.build_url(host_with_trailing_slash, path)



from dataclasses import field
from typing import Optional

from pydantic.dataclasses import dataclass
from pydantic_settings import SettingsConfigDict

DATABRICKS_ENV_PREFIX = "CASTOR_DATABRICKS_"


@dataclass
class DatabricksCredentials:
    """
    Credentials needed by Databricks client
    Requires:
    - host
    - token
    """

    host: str
    token: str = field(metadata={"sensitive": True})
    http_path: Optional[str] = field(default=None)

    model_config = SettingsConfigDict(
        env_prefix=DATABRICKS_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )



from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, Tuple

from requests.auth import HTTPBasicAuth


class CredentialsKey(Enum):
    """Value enum object for the credentials"""

    API_TOKEN = "api_token"  # noqa: S105
    BASE_URL = "base_url"
    CLIENT_ID = "client_id"
    CLOUD_ID = "cloud_id"
    DEVELOPER_TOKEN = "developer_token"  # noqa: S105


CLIENT_ALLOWED_KEYS: Tuple[str, ...] = tuple(c.value for c in CredentialsKey)


@dataclass
class DomoCredentials:
    """Class to handle Domo rest API permissions"""

    api_token: str
    base_url: str
    client_id: str
    developer_token: str
    cloud_id: Optional[str] = None

    @classmethod
    def from_secret(cls, secret: dict) -> "DomoCredentials":
        credentials = {
            k: v for k, v in secret.items() if k in CLIENT_ALLOWED_KEYS
        }
        return cls(**credentials)

    @property
    def authentication(self) -> HTTPBasicAuth:
        return HTTPBasicAuth(self.client_id, self.api_token)

    @property
    def private_headers(self) -> Dict[str, str]:
        return {"X-DOMO-Developer-Token": self.developer_token}



import logging
from typing import Iterable, Tuple

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_errors_logs,
    write_json,
    write_summary,
)
from .assets import TableauRevampAsset
from .client import TableauRevampClient

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: TableauRevampClient,
) -> Iterable[Tuple[TableauRevampAsset, list]]:
    """Iterate over the extracted Data from Tableau"""

    logger.info("Extracting USER from Tableau API")
    yield TableauRevampAsset.USER, deep_serialize(
        client.fetch(TableauRevampAsset.USER)
    )


def extract_all(client: TableauRevampClient, **kwargs: str) -> None:
    """
    Extract Data from tableau
    Store data locally in files under the output_directory
    If errors from Tableau's API are catch store them locally in file under the output_directory
    """
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)

    timestamp = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.value, output_directory, timestamp)
        write_json(filename, data)

    write_summary(
        output_directory,
        timestamp,
        base_url=client.base_url(),
        client_name=client.name(),
    )

    if client.errors:
        write_errors_logs(output_directory, timestamp, client.errors)



from dataclasses import dataclass, field
from typing import Optional

_DOMO_PUBLIC_URL = "https://api.domo.com"
_AUTH_URL = (
    "grant_type=client_credentials&scope=data%20dashboard%20audit%20user"
)


@dataclass
class Endpoint:
    """Contains endpoint URL and whether the API targeted is the private one"""

    base_url: str
    is_private: bool = False
    params: dict = field(default_factory=dict)

    def url(self, asset_id: Optional[str] = None):
        return f"{self.base_url}/{asset_id}" if asset_id else self.base_url


class EndpointFactory:
    """List of all Endpoints used for DOMO"""

    def __init__(self, base_url: str):
        self.base_url = base_url

    @property
    def authentication(self) -> Endpoint:
        return Endpoint(f"{_DOMO_PUBLIC_URL}/oauth/token?{_AUTH_URL}")

    @property
    def datasets(self) -> Endpoint:
        return Endpoint(f"{_DOMO_PUBLIC_URL}/v1/datasets")

    @property
    def pages(self) -> Endpoint:
        return Endpoint(f"{_DOMO_PUBLIC_URL}/v1/pages")

    @property
    def users(self) -> Endpoint:
        return Endpoint(f"{_DOMO_PUBLIC_URL}/v1/users")

    @property
    def dataflows(self) -> Endpoint:
        return Endpoint(
            base_url=f"{self.base_url}/api/dataprocessing/v1/dataflows",
            is_private=True,
        )

    def audit(self, start: int, end: int) -> Endpoint:
        """
        start and end are timestamps since epoch in milliseconds
        See [documentation](https://developer.domo.com/portal/a4e18ca6a0c0b-retrieve-activity-log-entries)
        """
        return Endpoint(
            base_url=f"{_DOMO_PUBLIC_URL}/v1/audit",
            params={"start": start, "end": end},
        )

    def lineage(self, dataflow_id: str) -> Endpoint:
        return Endpoint(
            base_url=f"{self.base_url}/api/data/v1/lineage/DATAFLOW/{dataflow_id}",
            is_private=True,
        )

    def table_lineage(self, dataset_id: str, cloud_id: str) -> Endpoint:
        return Endpoint(
            base_url=f"{self.base_url}/api/query/v1/byos/accounts/{cloud_id}/datasets/{dataset_id}/mapping",
            is_private=True,
        )

    def page_content(self, page_id: str) -> Endpoint:
        return Endpoint(
            base_url=f"{self.base_url}/api/content/v3/stacks/{page_id}/cards?parts=datasources",
            is_private=True,
        )



import logging
from typing import Iterator, Optional, Tuple

from requests import Response

from ...utils.client.api import APIClient
from .constants import DEFAULT_API_VERSION, DEFAULT_PAGINATION_LIMIT
from .credentials import SalesforceCredentials

logger = logging.getLogger(__name__)


class SalesforceBaseClient(APIClient):
    """
    Salesforce API client.
    https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/intro_rest.htm
    """

    api_version = DEFAULT_API_VERSION
    pagination_limit = DEFAULT_PAGINATION_LIMIT

    PATH_TPL = "services/data/v{version}/{suffix}"

    def __init__(self, credentials: SalesforceCredentials):
        super().__init__(host=credentials.base_url)
        self._token = self._access_token(credentials)

    def _access_token(self, credentials: SalesforceCredentials) -> str:
        url = self.build_url(self._host, "services/oauth2/token")
        response = self._call(
            url, "POST", params=credentials.token_request_payload()
        )
        return response["access_token"]

    def _full_url(self, suffix: str) -> str:
        path = self.PATH_TPL.format(version=self.api_version, suffix=suffix)
        return self.build_url(self._host, path)

    @property
    def query_url(self) -> str:
        """Returns the query API url"""
        return self._full_url("query")

    @property
    def tooling_url(self) -> str:
        """Returns the tooling API url"""
        return self._full_url("tooling/query")

    @staticmethod
    def _query_processor(response: Response) -> Tuple[dict, Optional[str]]:
        results = response.json()
        return results["records"], results.get("nextRecordsUrl")

    def _has_reached_pagination_limit(self, page_number: int) -> bool:
        return page_number > self.pagination_limit

    def _query_first_page(self, query: str) -> Tuple[Iterator[dict], str]:
        url = self.query_url
        logger.info("querying page 0")
        records, next_page_url = self._call(
            url, params={"q": query}, processor=self._query_processor
        )
        return records, next_page_url

    def _query_all(self, query: str) -> Iterator[dict]:
        """
        Run a SOQL query over salesforce API.

        more: https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm
        """
        records, next_page_path = self._query_first_page(query)
        yield from records

        page_count = 1
        while next_page_path and not self._has_reached_pagination_limit(
            page_count
        ):
            logger.info(f"querying page {page_count}")
            url = self.build_url(self._host, next_page_path)
            records, next_page_path = self._call(
                url, processor=self._query_processor
            )
            yield from records
            page_count += 1



from typing import Dict, Set

from ..assets import TableauRevampAsset

# list of fields to pick in TSC response
TSC_FIELDS: Dict[TableauRevampAsset, Set[str]] = {
    TableauRevampAsset.DATASOURCE: {
        "id",
        "project_id",
        "webpage_url",
    },
    TableauRevampAsset.PROJECT: {
        "description",
        "id",
        "name",
        "parent_id",
    },
    TableauRevampAsset.USAGE: {
        "name",
        "total_views",
        "workbook_id",
    },
    TableauRevampAsset.USER: {
        "email",
        "fullname",
        "id",
        "name",
        "site_role",
    },
    TableauRevampAsset.WORKBOOK: {
        "id",
        "project_id",
    },
}



import logging
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import date
from enum import Enum
from functools import partial
from typing import Any, Dict, List, Optional, Set, Tuple, cast

import requests
from databricks import sql  # type: ignore
from requests import Response

from ...utils import (
    SafeMode,
    at_midnight,
    date_after,
    mapping_from_rows,
    retry,
    safe_mode,
)
from ...utils.client.api import APIClient
from ...utils.pager import PagerOnToken
from ..abstract.time_filter import TimeFilter
from .credentials import DatabricksCredentials
from .format import DatabricksFormatter, TagMapping
from .types import Link, Ostr, OTimestampedLink, TablesColumns, TimestampedLink
from .utils import build_path, tag_label

logger = logging.getLogger(__name__)

_MAX_NUMBER_OF_LINEAGE_ERRORS = 1000
_MAX_THREADS = 10
_RETRY_ATTEMPTS = 3
_RETRY_BASE_MS = 1000
_RETRY_EXCEPTIONS = [
    requests.exceptions.ConnectTimeout,
]
_WORKSPACE_ID_HEADER = "X-Databricks-Org-Id"

_INFORMATION_SCHEMA_SQL = "SELECT * FROM system.information_schema"

safe_params = SafeMode((BaseException,), _MAX_NUMBER_OF_LINEAGE_ERRORS)


class TagEntity(Enum):
    """Entities that can be tagged in Databricks"""

    COLUMN = "COLUMN"
    TABLE = "TABLE"


def _day_to_epoch_ms(day: date) -> int:
    return int(at_midnight(day).timestamp() * 1000)


def _day_hour_to_epoch_ms(day: date, hour: int) -> int:
    return int(at_midnight(day).timestamp() * 1000) + (hour * 3600 * 1000)


class LineageLinks:
    """
    helper class that handles lineage deduplication and filtering
    """

    def __init__(self):
        self.lineage: Dict[Link, Ostr] = dict()

    def add(self, timestamped_link: TimestampedLink) -> None:
        """
        keep the most recent lineage link, adding to `self.lineage`
        """
        parent, child, timestamp = timestamped_link
        link = (parent, child)
        if not self.lineage.get(link):
            self.lineage[link] = timestamp
        else:
            if not timestamp:
                return
            # keep most recent link; cast for mypy
            recent = max(cast(str, self.lineage[link]), cast(str, timestamp))
            self.lineage[link] = recent


class DatabricksClient(APIClient):
    """Databricks Client"""

    def __init__(
        self,
        credentials: DatabricksCredentials,
        db_allowed: Optional[Set[str]] = None,
        db_blocked: Optional[Set[str]] = None,
        has_table_tags: bool = False,
        has_column_tags: bool = False,
    ):
        super().__init__(host=credentials.host, token=credentials.token)
        self._http_path = credentials.http_path
        self._db_allowed = db_allowed
        self._db_blocked = db_blocked
        self._has_table_tags = has_table_tags
        self._has_column_tags = has_column_tags
        self.formatter = DatabricksFormatter()

    def execute_sql(
        self,
        query: str,
        params: Optional[dict] = None,
    ):
        """
        Execute a SQL query on Databricks system tables and return the results.
        https://docs.databricks.com/en/dev-tools/python-sql-connector.html

        /!\ credentials.http_path is required in order to run SQL queries
        """
        assert self._http_path, "HTTP_PATH is required to run SQL queries"
        with sql.connect(
            server_hostname=self._host,
            http_path=self._http_path,
            access_token=self._token,
        ) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query, params)
                return cursor.fetchall()

    @staticmethod
    def name() -> str:
        return "Databricks"

    def _keep_catalog(self, catalog: str) -> bool:
        """
        Helper function to determine if we should keep the Databricks catalog
        which is a CastorDoc database
        """
        if self._db_allowed and catalog not in self._db_allowed:
            return False
        if self._db_blocked and catalog in self._db_blocked:
            return False
        return True

    def databases(self) -> List[dict]:
        path = "api/2.1/unity-catalog/catalogs"
        content = self.get(path=path)
        _databases = self.formatter.format_database(content.get("catalogs", []))
        return [d for d in _databases if self._keep_catalog(d["database_name"])]

    def _schemas_of_database(self, database: dict) -> List[dict]:
        path = "api/2.1/unity-catalog/schemas"
        payload = {"catalog_name": database["database_name"]}
        content = self.get(path=path, payload=payload)
        schemas = content.get("schemas", [])
        return self.formatter.format_schema(schemas, database)

    def schemas(self, databases: List[dict]) -> List[dict]:
        """
        Get the databricks schemas (also sometimes called databases)
        (which correspond to the schemas in Castor)
        leveraging the unity catalog API
        """
        return [
            schema
            for database in databases
            for schema in self._schemas_of_database(database)
        ]

    @staticmethod
    def _process_table_response(response: Response) -> Tuple[dict, str]:
        """
        Returns both the JSON content and the Workspace ID, which is found
        in the response's headers.
        """
        return response.json(), response.headers[_WORKSPACE_ID_HEADER]

    def _tables_columns_of_schema(
        self,
        schema: dict,
        table_tags: TagMapping,
        column_tags: TagMapping,
    ) -> TablesColumns:
        path = "api/2.1/unity-catalog/tables"
        payload = {
            "catalog_name": schema["database_id"],
            "schema_name": schema["schema_name"],
        }
        content, workspace_id = self.get(
            path=path,
            payload=payload,
            processor=self._process_table_response,
        )
        host = self.build_url(self._host, path="")
        return self.formatter.format_table_column(
            raw_tables=content.get("tables", []),
            schema=schema,
            host=host,
            workspace_id=workspace_id,
            table_tags=table_tags,
            column_tags=column_tags,
        )

    @staticmethod
    def _match_table_with_user(table: dict, user_mapping: dict) -> dict:
        table_owner_email = table.get("owner_email")
        if not table_owner_email:
            return table
        owner_external_id = user_mapping.get(table_owner_email)
        if not owner_external_id:
            return table
        return {**table, "owner_external_id": owner_external_id}

    def _needs_extraction(self, entity: TagEntity) -> bool:
        if entity == TagEntity.TABLE:
            return self._has_table_tags
        if entity == TagEntity.COLUMN:
            return self._has_column_tags
        raise AssertionError(f"Entity not supported: {entity}")

    def _get_tags_mapping(self, entity: TagEntity) -> TagMapping:
        """
        Fetch tags of the given entity and build a mapping:
        { path: list[tags] }

        https://docs.databricks.com/en/sql/language-manual/information-schema/table_tags.html
        https://docs.databricks.com/en/sql/language-manual/information-schema/column_tags.html
        """
        if not self._needs_extraction(entity):
            # extracting tags require additional credentials (http_path)
            return dict()

        table = f"{entity.value.lower()}_tags"
        query = f"{_INFORMATION_SCHEMA_SQL}.{table}"
        result = self.execute_sql(query)
        mapping = defaultdict(list)
        for row in result:
            dict_row = row.asDict()
            keys = ["catalog_name", "schema_name", "table_name"]
            if entity == TagEntity.COLUMN:
                keys.append("column_name")
            path = build_path(dict_row, keys)
            label = tag_label(dict_row)
            mapping[path].append(label)

        return mapping

    @staticmethod
    def _get_user_mapping(users: List[dict]) -> dict:
        return {
            **mapping_from_rows(users, "email", "id"),
            **mapping_from_rows(users, "user_name", "id"),
        }

    def tables_and_columns(
        self, schemas: List[dict], users: List[dict]
    ) -> TablesColumns:
        """
        Get the databricks tables & columns leveraging the unity catalog API
        """
        tables: List[dict] = []
        columns: List[dict] = []
        user_mapping = self._get_user_mapping(users)
        table_tags = self._get_tags_mapping(TagEntity.TABLE)
        column_tags = self._get_tags_mapping(TagEntity.COLUMN)
        for schema in schemas:

            t_to_add, c_to_add = self._tables_columns_of_schema(
                schema=schema,
                table_tags=table_tags,
                column_tags=column_tags,
            )
            t_with_owner = [
                self._match_table_with_user(table, user_mapping)
                for table in t_to_add
            ]
            tables.extend(t_with_owner)
            columns.extend(c_to_add)
        return tables, columns

    @staticmethod
    def _to_table_path(table: dict) -> Ostr:
        if table.get("name"):
            return f"{table['catalog_name']}.{table['schema_name']}.{table['name']}"
        return None

    @staticmethod
    def _to_column_path(column: dict) -> Ostr:
        if column.get("name"):
            return f"{column['catalog_name']}.{column['schema_name']}.{column['table_name']}.{column['name']}"
        return None

    def _link(
        self, path_from: Ostr, path_to: Ostr, timestamp: Ostr
    ) -> OTimestampedLink:
        """exclude missing path and self-lineage"""
        if (not path_from) or (not path_to):
            return None
        is_self_lineage = path_from.lower() == path_to.lower()
        if is_self_lineage:
            return None
        return (path_from, path_to, timestamp)

    def _single_table_lineage_links(
        self, table_path: str, single_table_lineage: dict
    ) -> List[TimestampedLink]:
        """
        process databricks lineage API response for a given table
        returns a list of (parent, child, timestamp)

        Note: in `upstreams` or `downstreams` we only care about `tableInfo`,
        we could also have `notebookInfos` or `fileInfo`
        """
        links: List[OTimestampedLink] = []
        # add parent:
        for link in single_table_lineage.get("upstreams", []):
            parent = link.get("tableInfo", {})
            parent_path = self._to_table_path(parent)
            timestamp: Ostr = parent.get("lineage_timestamp")
            links.append(self._link(parent_path, table_path, timestamp))

        # add children:
        for link in single_table_lineage.get("downstreams", []):
            child = link.get("tableInfo", {})
            child_path = self._to_table_path(child)
            timestamp = child.get("lineage_timestamp")
            links.append(self._link(table_path, child_path, timestamp))

        return list(filter(None, links))

    @safe_mode(safe_params, lambda: [])
    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_ATTEMPTS,
        base_ms=_RETRY_BASE_MS,
    )
    def get_single_table_lineage(
        self, table_path: str
    ) -> List[TimestampedLink]:
        """
        Helper function used in get_lineage_links.
        Call data lineage API and return the content of the result
        eg table_path: broward_prd.bronze.account_adjustments
        FYI: Maximum rate of 50 requests per SECOND
        """
        path = "api/2.0/lineage-tracking/table-lineage"
        payload = {"table_name": table_path, "include_entity_lineage": True}
        content = self.get(path=path, payload=payload)
        return self._single_table_lineage_links(table_path, content)

    def _deduplicate_lineage(self, lineages: List[TimestampedLink]) -> dict:
        deduplicated_lineage = LineageLinks()
        for timestamped_link in lineages:
            deduplicated_lineage.add(timestamped_link)
        return deduplicated_lineage.lineage

    def table_lineage(self, tables: List[dict]) -> List[dict]:
        """
        Wrapper function that retrieves all table lineage
        """
        # retrieve table lineage
        with ThreadPoolExecutor(max_workers=_MAX_THREADS) as executor:
            table_paths = [
                ".".join([table["schema_id"], table["table_name"]])
                for table in tables
            ]
            results = executor.map(self.get_single_table_lineage, table_paths)
        lineages = [link for links in results for link in links]
        deduplicated = self._deduplicate_lineage(lineages)
        return self.formatter.format_lineage(deduplicated)

    @staticmethod
    def _paths_for_column_lineage(
        tables: List[dict], columns: List[dict], table_lineage: List[dict]
    ) -> List[Tuple[str, str]]:
        """
        helper providing a list of candidate columns to look lineage for:
        we only look for column lineage where there is table lineage
        """
        # mapping between table id and its path db.schema.table
        # table["schema_id"] follows the pattern `db.schema`
        mapping = {
            table["id"]: ".".join([table["schema_id"], table["table_name"]])
            for table in tables
        }

        tables_with_lineage: Set[str] = set()
        for t in table_lineage:
            tables_with_lineage.add(t["parent_path"])
            tables_with_lineage.add(t["child_path"])

        paths_to_return: List[Tuple[str, str]] = []
        for column in columns:
            table_path = mapping[column["table_id"]]
            if table_path not in tables_with_lineage:
                continue
            column_ = (table_path, column["column_name"])
            paths_to_return.append(column_)

        return paths_to_return

    def _single_column_lineage_links(
        self, column_path: str, single_column_lineage: dict
    ) -> List[TimestampedLink]:
        """
        process databricks lineage API response for a given table
        returns a list of (parent, child, timestamp)

        Note: in `upstreams` or `downstreams` we only care about `tableInfo`,
        we could also have `notebookInfos` or `fileInfo`
        """
        links: List[OTimestampedLink] = []
        # add parent:
        for link in single_column_lineage.get("upstream_cols", []):
            parent_path = self._to_column_path(link)
            timestamp: Ostr = link.get("lineage_timestamp")
            links.append(self._link(parent_path, column_path, timestamp))

        # add children:
        for link in single_column_lineage.get("downstream_cols", []):
            child_path = self._to_column_path(link)
            timestamp = link.get("lineage_timestamp")
            links.append(self._link(column_path, child_path, timestamp))

        return list(filter(None, links))

    @safe_mode(safe_params, lambda: [])
    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_ATTEMPTS,
        base_ms=_RETRY_BASE_MS,
    )
    def get_single_column_lineage(
        self,
        names: Tuple[str, str],
    ) -> List[TimestampedLink]:
        """
        Helper function used in get_lineage_links.
        Call data lineage API and return the content of the result

        eg table_path: broward_prd.bronze.account_adjustments
        FYI: Maximum rate of 10 requests per SECOND
        """
        table_path, column_name = names
        api_path = "api/2.0/lineage-tracking/column-lineage"
        payload = {
            "table_name": table_path,
            "column_name": column_name,
            "include_entity_lineage": True,
        }
        content = self.get(path=api_path, payload=payload)
        column_path = f"{table_path}.{column_name}"
        return self._single_column_lineage_links(column_path, content)

    def column_lineage(
        self, tables: List[dict], columns: List[dict], table_lineage: List[dict]
    ) -> List[dict]:
        """
        Wrapper function that retrieves all column lineage
        we only try to retrieve column lineage if we found table lineage
        """
        candidate_paths = self._paths_for_column_lineage(
            tables, columns, table_lineage
        )
        lineages: List[TimestampedLink] = [
            link
            for paths in candidate_paths
            for link in self.get_single_column_lineage(paths)
        ]
        deduplicated = self._deduplicate_lineage(lineages)
        return self.formatter.format_lineage(deduplicated)

    @staticmethod
    def _time_filter(time_filter: Optional[TimeFilter]) -> dict:
        """time filter to retrieve Databricks' queries"""
        # define an explicit time window
        if not time_filter:
            time_filter = TimeFilter.default()

        assert time_filter  # for mypy

        hour_min = time_filter.hour_min
        hour_max = time_filter.hour_max
        day = time_filter.day
        if hour_min is not None and hour_max is not None:  # specific window
            start_time_ms = _day_hour_to_epoch_ms(day, hour_min)
            # note: in practice, hour_min == hour_max (hourly query ingestion)
            end_time_ms = _day_hour_to_epoch_ms(day, hour_max + 1)
        else:  # fallback to an extraction of the entire day
            start_time_ms = _day_to_epoch_ms(day)
            end_time_ms = _day_to_epoch_ms(date_after(day, 1))

        return {
            "filter_by": {
                "query_start_time_range": {
                    "end_time_ms": end_time_ms,
                    "start_time_ms": start_time_ms,
                }
            }
        }

    def query_payload(
        self,
        page_token: Optional[str] = None,
        max_results: Optional[int] = None,
        time_range_filter: Optional[dict] = None,
    ) -> dict:
        """helper method to build the payload used to retrieve queries"""
        # in payload: You can provide only one of 'page_token' or 'filter_by'
        if page_token:
            payload: Dict[str, Any] = {"page_token": page_token}
        else:
            if time_range_filter:
                payload = {**time_range_filter}
            else:
                payload = self._time_filter(None)  # default to yesterday
        if max_results:
            payload["max_results"] = max_results
        return payload

    def _scroll_queries(
        self,
        page_token: Optional[str] = None,
        max_results: Optional[int] = None,
        time_range_filter: Optional[dict] = None,
    ) -> dict:
        """
        Callback to scroll the queries api
        https://docs.databricks.com/api/workspace/queryhistory/list
        max_results: Limit the number of results returned in one page.
            The default is 100. (both on our side and Databricks')
        """
        path = "api/2.0/sql/history/queries"
        payload = self.query_payload(page_token, max_results, time_range_filter)
        content = self.get(path=path, payload=payload)
        return content if content else {}

    def queries(self, time_filter: Optional[TimeFilter] = None) -> List[dict]:
        """get all queries"""
        # add a time filter (by default: yesterday)
        time_range_filter = self._time_filter(time_filter)
        _time_filtered_scroll_queries = partial(
            self._scroll_queries,
            time_range_filter=time_range_filter,
        )

        # retrieve all queries using pagination
        raw_queries = PagerOnToken(_time_filtered_scroll_queries).all()

        return self.formatter.format_query(raw_queries)

    def users(self) -> List[dict]:
        """
        retrieve user from api
        """
        path = "api/2.0/preview/scim/v2/Users"
        content = self.get(path=path)
        return self.formatter.format_user(content.get("Resources", []))

    def _view_ddl(self, schema: dict) -> List[dict]:
        path = "api/2.1/unity-catalog/tables"
        payload = {
            "catalog_name": schema["database_id"],
            "schema_name": schema["schema_name"],
            "omit_columns": True,
        }
        content = self.get(path=path, payload=payload)
        return self.formatter.format_view_ddl(content.get("tables", []), schema)

    def view_ddl(self, schemas: List[dict]) -> List[dict]:
        """retrieve view ddl"""
        view_ddl: List[dict] = []
        for schema in schemas:
            v_to_add = self._view_ddl(schema)
            view_ddl.extend(v_to_add)
        return view_ddl



import logging
import os
from typing import Iterator

logger = logging.getLogger(__name__)

_ALLOWED_EXTENSION = (".json", ".csv")


def iter_files(repository_path: str) -> Iterator[str]:
    """
    Given a repository path yield all files in that repository
    Removes file whose extension is not allowed
    """

    for file in os.listdir(repository_path):
        _, ext = os.path.splitext(file)
        if ext not in _ALLOWED_EXTENSION:
            logger.info(f"Forbidden file extension : skipping {file}")
            continue
        file_path = os.path.join(repository_path, file)

        if os.path.isfile(file_path):
            yield file_path


def file_exist(f_path: str) -> bool:
    """
    Check if a file exist
    """
    return os.path.exists(os.path.expanduser(f_path))



# TSC for TableauServerClient: basic REST API to extracting core objects
from typing import Dict, Set

from .assets import TableauAsset

# TSC fields extracted per assets
TSC_FIELDS: Dict[TableauAsset, Set[str]] = {
    TableauAsset.PROJECT: {
        "id",
        "name",
        "description",
        "parent_id",
    },
    TableauAsset.PUBLISHED_DATASOURCE: {
        "id",
        "name",
        "description",
        "tags",
        "project_id",
        "created_at",
        "updated_at",
        "owner_id",
        "webpage_url",
    },
    TableauAsset.USAGE: {
        "workbook_id",
        "total_views",
    },
    TableauAsset.USER: {
        "id",
        "name",
        "email",
        "fullname",
        "site_role",
    },
    TableauAsset.WORKBOOK: {
        "id",
        "name",
        "description",
        "tags",
        "project_id",
        "created_at",
        "updated_at",
        "owner_id",
        "webpage_url",
    },
}



from typing import Callable, Iterator, Optional, Sequence, TypeVar
from uuid import UUID

from .pager import DEFAULT_PER_PAGE, AbstractPager, PagerStopStrategy

_DEFAULT_MIN_UUID = UUID("00000000-0000-0000-0000-000000000000")

T = TypeVar("T")


class PagerOnIdLogger:
    def on_iteration(self, max_id: UUID, count: int):
        pass

    def on_success(self, total: int):
        pass


class PagerOnId(AbstractPager):
    def __init__(
        self,
        callback: Callable[[UUID, int], Sequence[T]],
        *,
        logger: Optional[PagerOnIdLogger] = None,
        stop_strategy: PagerStopStrategy = PagerStopStrategy.EMPTY_PAGE,
    ):
        self._callback = callback
        self._logger = logger or PagerOnIdLogger()
        self._stop_strategy = stop_strategy

    @staticmethod
    def _max_id(items: Sequence) -> UUID:
        return max(item["id"] for item in items)

    def iterator(
        self,
        per_page: int = DEFAULT_PER_PAGE,
    ) -> Iterator[Sequence[T]]:
        """Yields data provided by the callback as a list using the greatest UUID as a reference point"""
        greater_than_id = _DEFAULT_MIN_UUID
        stop_on_empty_page = self._stop_strategy == PagerStopStrategy.EMPTY_PAGE

        total_results = 0

        while True:
            results = self._callback(greater_than_id, per_page)
            nb_results = len(results)
            total_results += nb_results

            if results:
                greater_than_id = self._max_id(results)
                self._logger.on_iteration(greater_than_id, nb_results)
                yield results

            stop = self.should_stop(nb_results, per_page, stop_on_empty_page)

            if stop:
                break
        self._logger.on_success(total_results)



from dataclasses import field

from pydantic.dataclasses import dataclass


@dataclass
class BigQueryCredentials:
    """Credentials needed by BigQuery client"""

    auth_provider_x509_cert_url: str
    auth_uri: str
    client_email: str
    client_id: str
    client_x509_cert_url: str
    private_key: str = field(metadata={"sensitive": True})
    private_key_id: str
    project_id: str
    token_uri: str
    type: str



from ...types import ExternalAsset


class TableauRevampAsset(ExternalAsset):
    """
    Tableau assets
    """

    COLUMN = "columns"
    DASHBOARD = "dashboards"
    DATASOURCE = "datasources"
    FIELD = "fields"
    PROJECT = "projects"
    SHEET = "sheets"
    TABLE = "tables"
    USAGE = "usage"
    USER = "users"
    WORKBOOK = "workbooks"



import logging
from typing import Dict, Optional

from ...utils import AbstractStorage, LocalStorage, write_summary
from ..abstract import (
    ADDITIONAL_LINEAGE_ASSETS,
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    SupportedAssets,
    TimeFilter,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
)
from .client import DatabricksClient
from .credentials import DatabricksCredentials

DATABRICKS_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.ADDITIONAL_LINEAGE: ADDITIONAL_LINEAGE_ASSETS,
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}

logger = logging.getLogger(__name__)

OTimeFilter = Optional[TimeFilter]
Paths = Dict[str, str]


class DatabricksExtractionProcessor:
    """Databricks' API-based extraction management"""

    def __init__(
        self,
        client: DatabricksClient,
        storage: AbstractStorage,
        skip_existing: bool = False,
    ):
        self._client = client
        self._storage = storage
        self._skip_existing = skip_existing

    def _should_not_reextract(self, asset_group: WarehouseAssetGroup) -> bool:
        """helper function to determine whether we need to extract"""
        if not self._skip_existing:
            return False

        for asset in DATABRICKS_ASSETS[asset_group]:
            if not self._storage.exists(asset.value):
                return False

        logger.info(f"Skipped, files for {asset_group.value} already exist")
        return True

    def _existing_group_paths(self, asset_group: WarehouseAssetGroup) -> Paths:
        return {
            a.value: self._storage.path(a.value)
            for a in DATABRICKS_ASSETS[asset_group]
        }

    def extract_catalog(self) -> Paths:
        """
        Extract all catalog assets (from database to column)
        and return the locations of the extracted data
        """
        if self._should_not_reextract(WarehouseAssetGroup.CATALOG):
            return self._existing_group_paths(WarehouseAssetGroup.CATALOG)

        catalog_locations: Dict[str, str] = dict()
        databases = self._client.databases()
        location = self._storage.put(WarehouseAsset.DATABASE.value, databases)
        catalog_locations[WarehouseAsset.DATABASE.value] = location
        logger.info(f"Extracted {len(databases)} databases to {location}")

        schemas = self._client.schemas(databases)
        location = self._storage.put(WarehouseAsset.SCHEMA.value, schemas)
        catalog_locations[WarehouseAsset.SCHEMA.value] = location
        logger.info(f"Extracted {len(schemas)} schemas to {location}")

        del databases

        users = self._client.users()
        tables, columns = self._client.tables_and_columns(schemas, users)

        location = self._storage.put(WarehouseAsset.TABLE.value, tables)
        catalog_locations[WarehouseAsset.TABLE.value] = location
        logger.info(f"Extracted {len(tables)} tables to {location}")

        location = self._storage.put(WarehouseAsset.COLUMN.value, columns)
        catalog_locations[WarehouseAsset.COLUMN.value] = location
        logger.info(f"Extracted {len(columns)} columns to {location}")
        return catalog_locations

    def extract_lineage(self) -> Paths:
        if self._should_not_reextract(WarehouseAssetGroup.ADDITIONAL_LINEAGE):
            return self._existing_group_paths(
                WarehouseAssetGroup.ADDITIONAL_LINEAGE
            )
        lineage_locations: Dict[str, str] = dict()

        # extract catalog
        databases = self._client.databases()
        schemas = self._client.schemas(databases)
        users = self._client.users()
        tables, columns = self._client.tables_and_columns(schemas, users)
        logger.info("Extracted pre-requisite catalog. Next comes lineage")

        # extract table lineage
        table_lineage = self._client.table_lineage(tables)
        table_lineage_key = WarehouseAsset.ADDITIONAL_TABLE_LINEAGE.value
        location = self._storage.put(table_lineage_key, table_lineage)
        lineage_locations[table_lineage_key] = location
        msg = f"Extracted {len(table_lineage)} table lineage to {location}"
        logger.info(msg)

        # extract column lineage
        column_lineage = self._client.column_lineage(
            tables, columns, table_lineage
        )
        column_lineage_key = WarehouseAsset.ADDITIONAL_COLUMN_LINEAGE.value
        location = self._storage.put(column_lineage_key, column_lineage)
        lineage_locations[column_lineage_key] = location
        msg = f"Extracted {len(column_lineage)} column lineage to {location}"
        logger.info(msg)
        return lineage_locations

    def extract_query(self, time_filter: OTimeFilter = None) -> Paths:
        """extract yesterday's queries and return their location"""
        if self._should_not_reextract(WarehouseAssetGroup.QUERY):
            return self._existing_group_paths(WarehouseAssetGroup.QUERY)

        queries = self._client.queries(time_filter)
        location = self._storage.put(WarehouseAsset.QUERY.value, queries)
        logger.info(f"Extracted {len(queries)} queries to {location}")
        return {WarehouseAsset.QUERY.value: location}

    def extract_role(self) -> Paths:
        """extract roles (users) and return their location"""
        if self._should_not_reextract(WarehouseAssetGroup.ROLE):
            return self._existing_group_paths(WarehouseAssetGroup.ROLE)

        users = self._client.users()
        location = self._storage.put(WarehouseAsset.USER.value, users)
        logger.info(f"Extracted {len(users)} users to {location}")
        return {WarehouseAsset.USER.value: location}

    def extract_view_ddl(self) -> Paths:
        """extract view ddl (using the same route as tables)"""
        if self._should_not_reextract(WarehouseAssetGroup.VIEW_DDL):
            return self._existing_group_paths(WarehouseAssetGroup.VIEW_DDL)

        databases = self._client.databases()
        schemas = self._client.schemas(databases)
        view_ddl = self._client.view_ddl(schemas)
        location = self._storage.put(WarehouseAsset.VIEW_DDL.value, view_ddl)
        logger.info(f"Extracted {len(view_ddl)} view_ddl to {location}")
        return {WarehouseAsset.VIEW_DDL.value: location}


def extract_all(**kwargs) -> None:
    """
    Extract all assets from Databricks and store the results in CSV files
    Time filter scope for `Queries` = the day before (from 12AM to 12AM)
    """
    output_directory, skip_existing = common_args(kwargs)

    client = DatabricksClient(
        credentials=DatabricksCredentials(**kwargs),
        db_allowed=kwargs.get("db_allowed"),
        db_blocked=kwargs.get("db_blocked"),
    )

    storage = LocalStorage(directory=output_directory)

    extractor = DatabricksExtractionProcessor(
        client=client,
        storage=storage,
        skip_existing=skip_existing,
    )

    extractor.extract_catalog()
    extractor.extract_lineage()
    extractor.extract_query()
    extractor.extract_role()
    extractor.extract_view_ddl()

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



from unittest.mock import Mock, patch

from .client import SigmaClient, SigmaCredentials

FAKE_CREDENTIALS = SigmaCredentials(  # noqa: S106
    host="IamFake",
    client_id="MeTwo",
    api_token="MeThree",
)


def _client() -> SigmaClient:
    return SigmaClient(credentials=FAKE_CREDENTIALS)


@patch("requests.get")
@patch.object(SigmaClient, "_get_headers")
def test__get(mocked_headers, mocked_request):
    client = _client()
    mocked_headers.return_value = {"header": "MyHeader"}
    fact = {
        "fact": "Lions are the only cats that live in groups, called prides. "
        "Every female within the pride is usually related.",
        "length": 109,
    }
    mocked_request.return_value = Mock(json=lambda: fact, status_code=200)

    result = client._get("https://catfact.ninja/fact")
    assert result == fact

    result = client._get("https://catfact.ninja/fact")["length"]
    assert result == 109

    mocked_request.return_value = Mock("not a json", status_code=200)

    result = client._get("https/whatev.er")
    assert result == {}


@patch.object(SigmaClient, "_get")
@patch.object(SigmaClient, "_get_headers")
def test__get_with_pagination(mocked_headers, mocked_request):
    client = _client()
    mocked_headers.return_value = {"header": "my_header"}
    mocked_request.return_value = {
        "nextPage": None,
        "entries": ["dataset_1", "dataset_2"],
        "total": 2,
    }

    assets = client._get_with_pagination("my_endpoint")
    assert list(assets) == ["dataset_1", "dataset_2"]

    mocked_request.assert_called_with("my_endpoint?page=0")



from abc import ABC
from enum import Enum
from typing import Literal, Set, TypedDict


class CsvOptions(TypedDict):
    delimiter: str
    quoting: Literal[1]
    quotechar: str


class classproperty(property):
    """
    Allow combination of @classmethod + @property
    https://stackoverflow.com/questions/5189699/how-to-make-a-class-property
    """

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class ExternalAsset(Enum):
    """
    Abstract class representing External Assets which can be extracted

    Enum's NAME identifies the asset.
        Example: TableauAsset.WORKBOOK

    Enum's VALUE gives the file name.
        Example: workbooks.json

    """

    __metaclass__ = ABC

    @classproperty
    def optional(cls) -> Set["ExternalAsset"]:
        """
        Returns the assets that are not necessarily extracted/pushed.
        Example:
        {
            WarehouseAsset.EXTERNAL_TABLE_LINEAGE,
            WarehouseAsset.EXTERNAL_COLUMN_LINEAGE,
        }
        """
        return set()

    @classproperty
    def mandatory(cls) -> Set["ExternalAsset"]:
        """
        Returns the assets that must always be provided.
        """
        return {
            item for item in cls if item not in cls.optional  # type: ignore
        }



from .credentials import SalesforceCredentials


def test_Credentials_token_request_payload():
    creds = SalesforceCredentials(
        username="giphy",
        password="1312",
        client_id="degenie",
        client_secret="fautpasledire",
        security_token="yo",
        base_url="man",
    )

    payload = creds.token_request_payload()

    assert payload["grant_type"] == "password"
    assert payload["username"] == "giphy"
    assert payload["client_id"] == "degenie"
    assert payload["client_secret"] == "fautpasledire"
    assert payload["password"] == "1312yo"



from unittest.mock import patch

from .client import SalesforceBaseClient
from .credentials import SalesforceCredentials


@patch.object(SalesforceBaseClient, "_call")
def test_SalesforceBaseClient__urls(mock_call):
    mock_call.return_value = {"access_token": "the_token"}
    credentials = SalesforceCredentials(
        username="usr",
        password="pw",
        client_id="key",
        client_secret="secret",
        security_token="token",
        base_url="url",
    )
    client = SalesforceBaseClient(credentials)

    assert client.query_url == "https://url/services/data/v59.0/query"
    assert client.tooling_url == "https://url/services/data/v59.0/tooling/query"



"""
Request timeout in seconds for Looker API
"""

DEFAULT_LOOKER_TIMEOUT_SECOND = 120
KEY_LOOKER_TIMEOUT_SECOND = "CASTOR_LOOKER_TIMEOUT_SECOND"

"""
Number of items per page when requesting Looker API
"""
DEFAULT_LOOKER_PAGE_SIZE = 500
KEY_LOOKER_PAGE_SIZE = "CASTOR_LOOKER_PAGE_SIZE"

"""
Maximum concurrent threads to run when fetching
"""
DEFAULT_LOOKER_THREAD_POOL_SIZE = 20
KEY_LOOKER_THREAD_POOL_SIZE = "CASTOR_LOOKER_THREAD_POOL_SIZE"

# env variables
BASE_URL = "CASTOR_LOOKER_BASE_URL"
CLIENT_ID = "CASTOR_LOOKER_CLIENT_ID"
CLIENT_SECRET = "CASTOR_LOOKER_CLIENT_SECRET"  # noqa: S105
SEARCH_PER_FOLDER = "CASTOR_LOOKER_SEARCH_PER_FOLDER"
LOG_TO_STDOUT = "CASTOR_LOOKER_LOG_TO_STDOUT"



import logging
from typing import Dict, List, Tuple

from ...utils import AbstractStorage, LocalStorage, write_summary
from ...utils.salesforce import to_credentials
from ..abstract import (
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
)
from .client import SalesforceClient

logger = logging.getLogger(__name__)


Paths = Dict[str, str]

SALESFORCE_CATALOG_ASSETS: Tuple[WarehouseAsset, ...] = (
    WarehouseAsset.TABLE,
    WarehouseAsset.COLUMN,
)

SALESFORCE_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.CATALOG: SALESFORCE_CATALOG_ASSETS
}


class SalesforceExtractionProcessor:
    """Salesforce API-based extraction management - warehouse part"""

    def __init__(
        self,
        client: SalesforceClient,
        storage: AbstractStorage,
        skip_existing: bool = False,
    ):
        self._client = client
        self._storage = storage
        self._skip_existing = skip_existing

    def _should_extract(self) -> bool:
        """helper function to determine whether we need to extract"""
        if not self._skip_existing:
            return True

        for asset in SALESFORCE_CATALOG_ASSETS:
            if not self._storage.exists(asset.value):
                return True

        logger.info("Skipped, files for catalog already exist")
        return False

    def _existing_group_paths(self) -> Paths:
        return {
            a.value: self._storage.path(a.value)
            for a in SALESFORCE_CATALOG_ASSETS
        }

    def extract_catalog(self, show_progress: bool = True) -> Paths:
        """
        Extract the following catalog assets: tables and columns
        and return the locations of the extracted data
        """
        if not self._should_extract():
            return self._existing_group_paths()

        catalog_locations: Paths = dict()

        tables = self._client.tables()
        location = self._storage.put(WarehouseAsset.TABLE.value, tables)
        catalog_locations[WarehouseAsset.TABLE.value] = location
        logger.info(f"Extracted {len(tables)} tables to {location}")

        sobject_names = [(t["api_name"], t["table_name"]) for t in tables]
        columns = self._client.columns(sobject_names, show_progress)
        location = self._storage.put(WarehouseAsset.COLUMN.value, columns)
        catalog_locations[WarehouseAsset.COLUMN.value] = location
        logger.info(f"Extracted {len(columns)} columns to {location}")
        return catalog_locations

    def extract_role(self) -> Paths:
        """extract no users and return the empty file location"""
        users: List[dict] = []
        location = self._storage.put(WarehouseAsset.USER.value, users)
        logger.info(f"Extracted {len(users)} users to {location}")
        return {WarehouseAsset.USER.value: location}


def extract_all(**kwargs) -> None:
    """
    Extract all assets from Salesforce and store the results in CSV files
    """
    output_directory, skip_existing = common_args(kwargs)

    client = SalesforceClient(credentials=to_credentials(kwargs))
    storage = LocalStorage(directory=output_directory)
    extractor = SalesforceExtractionProcessor(
        client=client,
        storage=storage,
        skip_existing=skip_existing,
    )

    extractor.extract_catalog()
    extractor.extract_role()

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



from typing import Dict, Tuple

from .format import (
    SCHEMA_NAME,
    SalesforceFormatter,
    _by_label,
    _field_description,
    _merge_label_and_api_name,
)


def _example_sobjects() -> Tuple[Dict[str, str], ...]:
    """Returns 4 sobjects with 2 sharing the same label"""
    a = {"Label": "a", "QualifiedApiName": "a_one"}
    b = {"Label": "b", "QualifiedApiName": "b"}
    c = {"Label": "c", "QualifiedApiName": "c"}
    a_prime = {"Label": "a", "QualifiedApiName": "a_two"}
    return a, b, c, a_prime


def test__field_description():

    field = {}
    assert _field_description(field) == ""

    definition = {}
    field = {"FieldDefinition": definition}
    assert _field_description(field) == ""

    definition.update({"Description": "foo"})
    assert "foo" in _field_description(field)

    field.update({"InlineHelpText": "bar"})
    assert "bar" in _field_description(field)

    definition.update({"ComplianceGroup": "bim"})
    assert "bim" in _field_description(field)

    definition.update({"SecurityClassification": "bam"})
    description = _field_description(field)

    assert "bam" in description
    expected = (
        "- Description: foo\n"
        "- Help Text: bar\n"
        "- Compliance Categorization: bim\n"
        "- Data Sensitivity Level: bam"
    )
    assert description == expected


def test__merge_label_and_api_name():
    sobject = {"Label": "foo", "QualifiedApiName": "bar"}
    payload = _merge_label_and_api_name(sobject)
    expected_name = "foo (bar)"
    assert payload == {
        "id": expected_name,
        "api_name": "bar",
        "label": "foo",
        "schema_id": SCHEMA_NAME,
        "table_name": expected_name,
        "description": "",
        "tags": [],
        "type": "TABLE",
    }


def test__by_label():
    a, b, c, a_prime = _example_sobjects()
    sobjects = [a, b, c, a_prime]
    by_label = _by_label(sobjects)
    assert by_label == {"a": [a, a_prime], "b": [b], "c": [c]}


def test_salesforce_formatter_tables():
    sobjects = [*_example_sobjects()]
    tables = SalesforceFormatter.tables(sobjects)
    expected_names = {"a (a_one)", "a (a_two)", "b", "c"}
    payload_names = {t["table_name"] for t in tables}
    assert payload_names == expected_names



from enum import Enum
from typing import Dict, Literal, Optional, overload

from ....utils import from_env

_AUTH_ERROR_MSG = "Need either user and password or token_name and token"

# To specify the default site on Tableau Server, you can use an empty string
# https://tableau.github.io/server-client-python/docs/api-ref#authentication
_DEFAULT_SERVER_SITE_ID = ""


class CredentialsKey(Enum):
    """Value enum object for the credentials"""

    TABLEAU_USER = "user"
    TABLEAU_PASSWORD = "password"  # noqa: S105
    TABLEAU_TOKEN_NAME = "token_name"  # noqa: S105
    TABLEAU_TOKEN = "token"  # noqa: S105
    TABLEAU_SITE_ID = "site_id"
    TABLEAU_SERVER_URL = "server_url"


CREDENTIALS_ENV: Dict[CredentialsKey, str] = {
    CredentialsKey.TABLEAU_USER: "CASTOR_TABLEAU_USER",
    CredentialsKey.TABLEAU_PASSWORD: "CASTOR_TABLEAU_PASSWORD",
    CredentialsKey.TABLEAU_TOKEN_NAME: "CASTOR_TABLEAU_TOKEN_NAME",
    CredentialsKey.TABLEAU_TOKEN: "CASTOR_TABLEAU_TOKEN",
    CredentialsKey.TABLEAU_SITE_ID: "CASTOR_TABLEAU_SITE_ID",
    CredentialsKey.TABLEAU_SERVER_URL: "CASTOR_TABLEAU_SERVER_URL",
}


@overload
def get_value(key: CredentialsKey, kwargs: dict) -> Optional[str]: ...


@overload
def get_value(
    key: CredentialsKey, kwargs: dict, optional: Literal[True]
) -> Optional[str]: ...


@overload
def get_value(
    key: CredentialsKey, kwargs: dict, optional: Literal[False]
) -> str: ...


def get_value(
    key: CredentialsKey,
    kwargs: dict,
    optional: bool = True,
) -> Optional[str]:
    """
    Returns the value of the given key:
    - from kwargs in priority
    - from ENV otherwise
    Raises an error if not found (unless optional)
    """

    if key.value in kwargs:
        return kwargs[key.value]

    env_key = CREDENTIALS_ENV[key]
    return from_env(env_key, optional)


class TableauRevampCredentials:
    """
    Tableau's credentials to connect to REST API
    """

    def __init__(
        self,
        *,
        server_url: str,
        site_id: Optional[str],
        user: Optional[str],
        password: Optional[str],
        token_name: Optional[str],
        token: Optional[str],
    ):
        self.user = user
        self.site_id = site_id or _DEFAULT_SERVER_SITE_ID
        self.server_url = server_url
        self.password = password
        self.token_name = token_name
        self.token = token

    @classmethod
    def from_env(cls, kwargs: dict) -> "TableauRevampCredentials":
        return TableauRevampCredentials(
            server_url=get_value(
                CredentialsKey.TABLEAU_SERVER_URL,
                kwargs,
                optional=False,
            ),
            site_id=get_value(CredentialsKey.TABLEAU_SITE_ID, kwargs),
            user=get_value(CredentialsKey.TABLEAU_USER, kwargs),
            password=get_value(CredentialsKey.TABLEAU_PASSWORD, kwargs),
            token_name=get_value(CredentialsKey.TABLEAU_TOKEN_NAME, kwargs),
            token=get_value(CredentialsKey.TABLEAU_TOKEN, kwargs),
        )



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import metabase  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    # mandatory
    parser.add_argument(
        "-H",
        "--host",
        help="Host name where the server is running",
    )
    parser.add_argument("-P", "--port", help="TCP/IP port number")
    parser.add_argument("-d", "--database", help="Database name")
    parser.add_argument("-s", "--schema", help="Schema name")
    parser.add_argument("-u", "--username", help="Username")
    parser.add_argument("-p", "--password", help="Password")
    parser.add_argument(
        "-k",
        "--encryption_secret_key",
        help="Encryption secret key",
    )
    parser.add_argument(
        "--require_ssl",
        action="store_true",
        help="Require SSL",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()

    client = metabase.DbClient(
        host=args.host,
        port=args.port,
        database=args.database,
        schema=args.schema,
        user=args.username,
        password=args.password,
        encryption_secret_key=args.encryption_secret_key,
        require_ssl=args.require_ssl,
    )

    metabase.extract_all(
        client,
        output_directory=args.output,
    )



from datetime import datetime, timedelta

import pytest

from .constants import Assertions
from .utils import batch_size_is_valid_or_assert, datetime_is_recent_or_assert


def test_batch_size_is_valid_or_assert():
    valid = [1, 3, 4]
    batch_size_is_valid_or_assert(valid)

    invalid = list(range(8000))
    with pytest.raises(AssertionError, match=Assertions.BATCH_TOO_BIG):
        batch_size_is_valid_or_assert(invalid)


def test_datetime_is_recent_or_assert():
    krach = datetime(1929, 10, 29)
    with pytest.raises(AssertionError, match=Assertions.DATETIME_TOO_OLD):
        datetime_is_recent_or_assert(krach)

    yesterday = datetime.today() - timedelta(1)
    datetime_is_recent_or_assert(yesterday)



import logging
import random
import time
from enum import Enum
from typing import Any, Callable, Sequence, Tuple, Type, Union

from pydantic import BaseModel, PositiveInt, PrivateAttr
from pydantic.fields import Field

logger = logging.getLogger(__name__)


MS_IN_SEC = 1000


class RetryStrategy(Enum):
    """retry mechanism"""

    CONSTANT = "CONSTANT"
    LINEAR = "LINEAR"
    EXPONENTIAL = "EXPONENTIAL"


DEFAULT_STRATEGY = RetryStrategy.CONSTANT


class Retry(BaseModel):
    """
    This class checks if the retry conditions are met, and if so, how long to
    wait until the next attempt.
    """

    max_retries: PositiveInt
    base_ms: PositiveInt
    jitter_ms: PositiveInt
    strategy: RetryStrategy = Field(DEFAULT_STRATEGY, kw_only=True)
    _retry_attempts = PrivateAttr(default=0)

    def jitter(self) -> int:
        """compute random jitter in ms"""
        min_ = self.jitter_ms // 2
        max_ = int(self.jitter_ms * 1.5)
        return random.randrange(min_, max_)

    def base(self) -> int:
        """compute base wait time in ms"""
        base = self.base_ms
        if self.strategy == RetryStrategy.CONSTANT:
            return base
        if self.strategy == RetryStrategy.LINEAR:
            return base * self._retry_attempts
        # exponential
        scaled = float(base) / MS_IN_SEC
        return int(scaled**self._retry_attempts * MS_IN_SEC)

    def check(self, error: BaseException, log_exc_info: bool = False) -> bool:
        """
        Check whether retry should happen or not.
        If log_exc_info is True, it will add a more extensive log (most notably
        including the traceback).
        """
        if self._retry_attempts >= self.max_retries:
            return False

        exc_info = error if log_exc_info else None
        logger.warning("Caught a retryable exception", exc_info=exc_info)

        self._retry_attempts += 1
        wait_ms = self.base() + self.jitter()
        wait_s = float(wait_ms) / MS_IN_SEC
        msg = f"Attempting a new call in {wait_s} seconds, {self._retry_attempts} attempt(s) / {self.max_retries} max retries"
        logger.warning(msg)
        time.sleep(wait_s)
        return True


WrapperReturnType = Union[Tuple[BaseException, None], Tuple[None, Any]]


def retry(
    exceptions: Sequence[Type[BaseException]],
    max_retries: int = 1,
    base_ms: int = 0,
    jitter_ms: int = 1,
    strategy: RetryStrategy = DEFAULT_STRATEGY,
    log_exc_info: bool = False,
) -> Callable:
    """retry decorator"""

    exceptions_ = tuple(e for e in exceptions)

    def _wrapper(callable: Callable) -> Callable:
        def _try(*args, **kwargs) -> WrapperReturnType:
            try:
                return None, callable(*args, **kwargs)
            except exceptions_ as err:
                logger.warning(f"Exception within {callable.__name__}")
                return err, None

        def _func(*args, **kwargs) -> Any:
            retry = Retry(
                max_retries=max_retries,
                base_ms=base_ms,
                jitter_ms=jitter_ms,
                strategy=strategy,
            )
            while True:
                err, result = _try(*args, **kwargs)
                if err is None:
                    return result
                if retry.check(err, log_exc_info):
                    continue
                raise err

        return _func

    return _wrapper



import logging
from typing import Dict, Iterator, List, Optional

import tableauserverclient as TSC  # type: ignore
from tableauserverclient import Pager

from ....utils import SerializedAsset
from ..assets import TableauRevampAsset
from ..constants import (
    DEFAULT_PAGE_SIZE,
    DEFAULT_TIMEOUT_SECONDS,
    TABLEAU_SERVER_VERSION,
)
from .credentials import TableauRevampCredentials
from .errors import TableauApiError
from .gql_queries import FIELDS_QUERIES, GQL_QUERIES, QUERY_TEMPLATE
from .tsc_fields import TSC_FIELDS

logger = logging.getLogger(__name__)

# these assets must be extracted via TableauServerClient
_TSC_ASSETS = (
    # only users who published content can be extracted from MetadataAPI
    TableauRevampAsset.USER,
    # projects are not available in Metadata API
    TableauRevampAsset.PROJECT,
    # view count are not available in Metadata API
    TableauRevampAsset.USAGE,
)

# increase the value when extraction is too slow
# decrease the value when timeouts arise
_CUSTOM_PAGE_SIZE: Dict[TableauRevampAsset, int] = {
    # for some clients, extraction of columns tend to hit the node limit
    # https://community.tableau.com/s/question/0D54T00000YuK60SAF/metadata-query-nodelimitexceeded-error
    # the workaround is to reduce pagination
    TableauRevampAsset.COLUMN: 50,
    # fields are light but volumes are bigger
    TableauRevampAsset.FIELD: 1000,
    TableauRevampAsset.TABLE: 50,
}


def _pick_fields(
    data: Pager,
    asset: TableauRevampAsset,
) -> SerializedAsset:
    fields = TSC_FIELDS[asset]

    def _pick(row: dict):
        return {field: getattr(row, field) for field in fields}

    return [_pick(row) for row in data]


def _enrich_datasources_with_tsc(
    datasources: SerializedAsset,
    tsc_datasources: SerializedAsset,
) -> SerializedAsset:
    """
    Enrich datasources with fields coming from TableauServerClient:
    - project_luid
    - webpage_url
    """

    mapping = {row["id"]: row for row in tsc_datasources}

    for datasource in datasources:
        if datasource["__typename"] != "PublishedDatasource":
            # embedded datasources are bound to workbooks => no project
            # embedded datasources cannot be accessed via URL => no webpage_url
            continue
        luid = datasource["luid"]
        tsc_datasource = mapping[luid]
        datasource["projectLuid"] = tsc_datasource["project_id"]
        datasource["webpageUrl"] = tsc_datasource["webpage_url"]

    return datasources


def _enrich_workbooks_with_tsc(
    workbooks: SerializedAsset,
    tsc_workbooks: SerializedAsset,
) -> SerializedAsset:
    """
    Enrich workbooks with fields coming from TableauServerClient:
    - project_luid
    """

    mapping = {row["id"]: row for row in tsc_workbooks}

    for workbook in workbooks:
        luid = workbook["luid"]
        tsc_workbook = mapping.get(luid)
        if not tsc_workbook:
            # it happens that a workbook is in Metadata API but not in TSC
            # in this case, we push the workbook with default project
            logger.warning(f"Workbook {luid} was not found in TSC")
            workbook["projectLuid"] = None
            continue

        workbook["projectLuid"] = tsc_workbook["project_id"]

    return workbooks


def gql_query_scroll(
    server,
    query: str,
    resource: str,
) -> Iterator[SerializedAsset]:
    """Iterate over GQL query results, handling pagination and cursor"""

    def _call(cursor: Optional[str]) -> dict:
        # If cursor is defined it must be quoted else use null token
        token = "null" if cursor is None else f'"{cursor}"'
        query_ = query.replace("AFTER_TOKEN_SIGNAL", token)
        answer = server.metadata.query(query_)
        if "errors" in answer:
            raise TableauApiError(answer["errors"])
        return answer["data"][f"{resource}Connection"]

    cursor = None
    while True:
        payload = _call(cursor)
        yield payload["nodes"]

        page_info = payload["pageInfo"]
        if page_info["hasNextPage"]:
            cursor = page_info["endCursor"]
        else:
            break


class TableauRevampClient:
    """
    Connect to Tableau's API and extract assets.

    Relies on TableauServerClient overlay:
    https://tableau.github.io/server-client-python/docs/
    - for connection
    - to extract Users (Metadata

    Calls the MetadataAPI, using graphQL
    https://help.tableau.com/current/api/metadata_api/en-us/reference/index.html
    """

    def __init__(
        self,
        credentials: TableauRevampCredentials,
        timeout_sec: int = DEFAULT_TIMEOUT_SECONDS,
    ):
        self._credentials = credentials
        self._server = TSC.Server(self._credentials.server_url)
        options = {"verify": True, "timeout": timeout_sec}
        self._server.add_http_options(options)
        self._server.version = TABLEAU_SERVER_VERSION
        self.errors: List[str] = []

    @staticmethod
    def name() -> str:
        return "Tableau/API"

    def _user_password_login(self) -> None:
        """Login into Tableau using user and password"""
        self._server.auth.sign_in(
            TSC.TableauAuth(
                self._credentials.user,
                self._credentials.password,
                site_id=self._credentials.site_id,
            ),
        )

    def _pat_login(self) -> None:
        """Login into Tableau using personal authentication token"""
        self._server.auth.sign_in(
            TSC.PersonalAccessTokenAuth(
                self._credentials.token_name,
                self._credentials.token,
                site_id=self._credentials.site_id,
            ),
        )

    def login(self) -> None:
        """
        Depending on the given credentials, logs-in using either:
        - user/password
        - token_name/value (Personal Access Token)
        https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_concepts_auth.htm

        Raises an error if none can be found
        """

        if self._credentials.user and self._credentials.password:
            logger.info("Logging in using user and password authentication")
            return self._user_password_login()

        if self._credentials.token_name and self._credentials.token:
            logger.info("Logging in using token authentication")
            return self._pat_login()

        raise ValueError(
            "Invalid credentials: either user/password or PAT must be provided",
        )

    def base_url(self) -> str:
        return self._credentials.server_url

    def _fetch_from_tsc(
        self,
        asset: TableauRevampAsset,
    ) -> SerializedAsset:

        if asset == TableauRevampAsset.DATASOURCE:
            data = TSC.Pager(self._server.datasources)

        elif asset == TableauRevampAsset.PROJECT:
            data = TSC.Pager(self._server.projects)

        elif asset == TableauRevampAsset.USAGE:
            data = TSC.Pager(self._server.views, usage=True)

        elif asset == TableauRevampAsset.USER:
            data = TSC.Pager(self._server.users)

        elif asset == TableauRevampAsset.WORKBOOK:
            data = TSC.Pager(self._server.workbooks)

        else:
            raise AssertionError(f"Fetching from TSC not supported for {asset}")

        return _pick_fields(data, asset)

    def _run_graphql_query(
        self,
        resource: str,
        fields: str,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> SerializedAsset:
        query = QUERY_TEMPLATE.format(
            resource=resource,
            fields=fields,
            page_size=page_size,
        )
        result_pages = gql_query_scroll(self._server, query, resource)
        return [asset for page in result_pages for asset in page]

    def _fetch_fields(self) -> SerializedAsset:
        result: SerializedAsset = []
        page_size = _CUSTOM_PAGE_SIZE[TableauRevampAsset.FIELD]
        for resource, fields in FIELDS_QUERIES:
            current = self._run_graphql_query(resource, fields, page_size)
            result.extend(current)
        return result

    def _fetch_from_metadata_api(
        self,
        asset: TableauRevampAsset,
    ) -> SerializedAsset:
        if asset == TableauRevampAsset.FIELD:
            return self._fetch_fields()

        page_size = _CUSTOM_PAGE_SIZE.get(asset) or DEFAULT_PAGE_SIZE
        resource, fields = GQL_QUERIES[asset]
        return self._run_graphql_query(resource, fields, page_size)

    def _fetch_datasources(self) -> SerializedAsset:
        asset = TableauRevampAsset.DATASOURCE

        datasources = self._fetch_from_metadata_api(asset)
        datasource_projects = self._fetch_from_tsc(asset)

        return _enrich_datasources_with_tsc(datasources, datasource_projects)

    def _fetch_workbooks(self) -> SerializedAsset:
        asset = TableauRevampAsset.WORKBOOK

        workbooks = self._fetch_from_metadata_api(asset)
        workbook_projects = self._fetch_from_tsc(asset)

        return _enrich_workbooks_with_tsc(workbooks, workbook_projects)

    def fetch(
        self,
        asset: TableauRevampAsset,
    ) -> SerializedAsset:
        """
        Extract the given Tableau Asset
        """
        if asset == TableauRevampAsset.DATASOURCE:
            # both APIs are required to extract datasources
            return self._fetch_datasources()

        if asset == TableauRevampAsset.WORKBOOK:
            # both APIs are required to extract workbooks
            return self._fetch_workbooks()

        if asset in _TSC_ASSETS:
            # some assets can only be extracted via TSC
            return self._fetch_from_tsc(asset)

        # extract most assets via Metadata API
        return self._fetch_from_metadata_api(asset)



from enum import Enum
from typing import Dict, List, Set, Tuple

from ...types import ExternalAsset, classproperty


class WarehouseAsset(ExternalAsset):
    """Assets that can be extracted from warehouses"""

    ADDITIONAL_COLUMN_LINEAGE = "additional_column_lineage"
    ADDITIONAL_TABLE_LINEAGE = "additional_table_lineage"
    COLUMN = "column"
    COLUMN_LINEAGE = "column_lineage"  # specific to snowflake
    DATABASE = "database"
    EXTERNAL_COLUMN_LINEAGE = "external_column_lineage"
    EXTERNAL_TABLE_LINEAGE = "external_table_lineage"
    GRANT_TO_ROLE = "grant_to_role"
    GRANT_TO_USER = "grant_to_user"
    GROUP = "group"
    QUERY = "query"
    ROLE = "role"
    SCHEMA = "schema"
    TABLE = "table"
    FUNCTION = "function"
    USER = "user"
    VIEW_DDL = "view_ddl"

    @classproperty
    def optional(cls) -> Set["WarehouseAsset"]:
        return {
            WarehouseAsset.ADDITIONAL_COLUMN_LINEAGE,
            WarehouseAsset.ADDITIONAL_TABLE_LINEAGE,
            WarehouseAsset.EXTERNAL_COLUMN_LINEAGE,
            WarehouseAsset.EXTERNAL_TABLE_LINEAGE,
            WarehouseAsset.FUNCTION,
        }


class WarehouseAssetGroup(Enum):
    """Groups of assets that can be extracted together"""

    ADDITIONAL_LINEAGE = "additional_lineage"
    CATALOG = "catalog"
    EXTERNAL_LINEAGE = "external_lineage"
    FUNCTION = "function"
    QUERY = "query"
    ROLE = "role"
    SNOWFLAKE_LINEAGE = "snowflake_lineage"
    VIEW_DDL = "view_ddl"


# tuple of supported assets for each group (depends on the technology)
SupportedAssets = Dict[WarehouseAssetGroup, Tuple[WarehouseAsset, ...]]

# shared by all technologies
CATALOG_ASSETS = (
    WarehouseAsset.DATABASE,
    WarehouseAsset.SCHEMA,
    WarehouseAsset.TABLE,
    WarehouseAsset.COLUMN,
)

# shared by technologies supporting queries
FUNCTIONS_ASSETS = (WarehouseAsset.FUNCTION,)
QUERIES_ASSETS = (WarehouseAsset.QUERY,)
VIEWS_ASSETS = (WarehouseAsset.VIEW_DDL,)

EXTERNAL_LINEAGE_ASSETS = (
    WarehouseAsset.EXTERNAL_COLUMN_LINEAGE,
    WarehouseAsset.EXTERNAL_TABLE_LINEAGE,
)

ADDITIONAL_LINEAGE_ASSETS = (
    WarehouseAsset.ADDITIONAL_COLUMN_LINEAGE,
    WarehouseAsset.ADDITIONAL_TABLE_LINEAGE,
)

NON_EXTRACTABLE_ASSETS = {WarehouseAssetGroup.EXTERNAL_LINEAGE}


def extractable_asset_groups(
    supported_assets: SupportedAssets,
) -> List[Tuple[WarehouseAsset, ...]]:
    """
    helper function to differentiate
    extractable assets vs supported (ingest-able) assets
    """
    groups = set(supported_assets).difference(NON_EXTRACTABLE_ASSETS)
    extractable: Set[Tuple[WarehouseAsset, ...]] = {
        supported_assets[group] for group in groups
    }
    return list(extractable)



SOBJECTS_QUERY_TPL = """
    SELECT
        DeveloperName,
        DurableId,
        ExternalSharingModel,
        InternalSharingModel,
        Label,
        PluralLabel,
        QualifiedApiName
    FROM EntityDefinition
    WHERE DurableId > '{start_durable_id}'
    ORDER BY DurableId
    LIMIT {limit}
"""


SOBJECT_FIELDS_QUERY_TPL = """
    SELECT
        DataType,
        DeveloperName,
        Digits,
        FieldDefinition.BusinessOwnerId,
        FieldDefinition.ComplianceGroup,
        FieldDefinition.DataType,
        FieldDefinition.Description,
        FieldDefinition.IsIndexed,
        FieldDefinition.LastModifiedBy.Username,
        FieldDefinition.LastModifiedDate,
        FieldDefinition.SecurityClassification,
        InlineHelpText,
        IsComponent,
        IsCompound,
        IsNillable,
        IsUnique,
        Label,
        Length,
        Precision,
        QualifiedApiName,
        ReferenceTo,
        RelationshipName,
        Scale
    FROM EntityParticle
    WHERE EntityDefinitionId='{entity_definition_id}'
"""



from .asset import (
    EXTERNAL_LINEAGE_ASSETS,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    extractable_asset_groups,
)

COOL_TECHNOLOGY_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.CATALOG: (
        WarehouseAsset.DATABASE,
        WarehouseAsset.SCHEMA,
    ),
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}


def test_extractable_asset_groups():
    all_assets = set()
    for group in extractable_asset_groups(COOL_TECHNOLOGY_ASSETS):
        all_assets.update(group)

    expected = {
        WarehouseAsset.DATABASE,
        WarehouseAsset.SCHEMA,
        WarehouseAsset.USER,
    }  # no external lineage
    assert all_assets == expected



import logging
from itertools import chain
from typing import Callable, Iterator, List, Optional, Tuple

from ...utils import (
    OUTPUT_DIR,
    AbstractSourceClient,
    AbstractStorage,
    SafeMode,
    from_env,
    safe_mode,
)
from .asset import WarehouseAsset
from .query import AbstractQueryBuilder, ExtractionQuery

logger = logging.getLogger(__name__)


def common_args(kwargs: dict) -> Tuple[str, bool]:
    """Args used by all technologies"""
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)
    skip_existing = kwargs.get("skip_existing") or False
    return output_directory, skip_existing


class SQLExtractionProcessor:
    """extraction management"""

    def __init__(
        self,
        client: AbstractSourceClient,
        query_builder: AbstractQueryBuilder,
        storage: AbstractStorage,
        safe_mode: Optional[SafeMode] = None,
    ):
        self._client = client
        self._query_builder = query_builder
        self._storage = storage
        self._safe_mode = safe_mode

    @staticmethod
    def _unique(data: Iterator[dict]) -> List[dict]:
        """
        Remove duplicate in the given data.
        Remark: this method implies loading all data in memory: it breaks the streaming pipeline !
        """
        # dict > set > dict
        return [dict(t) for t in {tuple(d.items()) for d in data}]

    def _fetch(self, query: ExtractionQuery) -> Iterator[dict]:
        default: Callable[[], Iterator] = lambda: iter(())  # type: ignore
        decorator = safe_mode(self._safe_mode, default)
        decorated_execute = decorator(self._client.execute)
        return decorated_execute(query)

    def _results(self, asset: WarehouseAsset) -> Iterator[dict]:
        data: Iterator[dict] = iter([])
        queries = self._query_builder.build(asset)
        total = len(queries)

        for i, query in enumerate(queries):
            logger.info(f"Extracting {asset.value}: query {i+1}/{total}")
            # concatenate results of all queries
            data = chain(data, self._fetch(query))

        if self._query_builder.needs_deduplication(asset):
            # cast the list to iterator, but the streaming pipeline is broken in that case
            return (row for row in self._unique(data))

        return data

    def extract(
        self,
        asset: WarehouseAsset,
        skip_existing: bool = False,
    ) -> str:
        """
        Process extraction for the given asset and returns the location of extracted data
        """
        asset_name = asset.value
        if skip_existing and self._storage.exists(asset_name):
            logger.info("Skipped, file already exists")
            return self._storage.path(asset_name)

        try:
            data = self._results(asset)
            return self._storage.put(asset_name, data)
        finally:
            self._client.close(dispose=True)



import json

from ....utils import load_file
from ..assets import EXPORTED_FIELDS, ModeAnalyticsAsset
from .client import Client

_HOST = "https://mode.com"
_WORKSPACE = "castor"


def _dummy_client() -> Client:
    return Client(  # noqa: S106
        host=_HOST,
        workspace=_WORKSPACE,
        token="dummy-token",
        secret="******",
        no_checks=True,
    )


def test__url():
    client = _dummy_client()
    # basic calls
    value_1 = client._url(resource_name="spaces")
    assert value_1 == f"{_HOST}/api/{_WORKSPACE}/spaces"

    # with spaces
    value_2 = client._url(space="123456789", resource_name="reports")
    assert value_2 == f"{_HOST}/api/{_WORKSPACE}/spaces/123456789/reports"

    # with report
    value_3 = client._url(report="xxx", resource_name="queries")
    assert value_3 == f"{_HOST}/api/{_WORKSPACE}/reports/xxx/queries"

    # without workspace
    value_4 = client._url(resource_name="john_doe", with_workspace=False)
    assert value_4 == f"{_HOST}/api/john_doe"


def test__post_processing():
    client = _dummy_client()
    raw = load_file("client_test.json", __file__)
    result = client._post_processing(
        asset=ModeAnalyticsAsset.COLLECTION,
        data=[json.loads(raw)],
    )
    collection = result[0]
    assert set(collection.keys()) == set(
        EXPORTED_FIELDS[ModeAnalyticsAsset.COLLECTION],
    )
    assert collection["creator"] == "john_doe"



import logging
from typing import Dict, Iterator, List, Optional

from ....utils.salesforce import SalesforceBaseClient
from ..assets import SalesforceReportingAsset
from .soql import queries

logger = logging.getLogger(__name__)

REQUIRING_URL_ASSETS = (
    SalesforceReportingAsset.REPORTS,
    SalesforceReportingAsset.DASHBOARDS,
    SalesforceReportingAsset.FOLDERS,
)


class SalesforceReportingClient(SalesforceBaseClient):
    """
    Salesforce Reporting API client
    """

    def _get_asset_url(
        self, asset_type: SalesforceReportingAsset, asset: dict
    ) -> Optional[str]:
        """
        Fetch the given Asset + add the corresponding URL.
        """

        if asset_type == SalesforceReportingAsset.DASHBOARDS:
            path = f"lightning/r/Dashboard/{asset['Id']}/view"
            return self.build_url(self._host, path)

        if asset_type == SalesforceReportingAsset.FOLDERS:
            path = asset["attributes"]["url"].lstrip("/")
            return self.build_url(self._host, path)

        if asset_type == SalesforceReportingAsset.REPORTS:
            path = f"lightning/r/Report/{asset['Id']}/view"
            return self.build_url(self._host, path)

        return None

    def _fetch_and_add_url(
        self, asset_type: SalesforceReportingAsset
    ) -> Iterator[dict]:
        assets = self._query_all(queries[asset_type])
        for asset in assets:
            url = self._get_asset_url(asset_type, asset)
            yield {**asset, "Url": url}

    def fetch(self, asset: SalesforceReportingAsset) -> List[Dict]:
        """
        Fetch Salesforce Reporting assets
        """
        logger.info(f"Starting extraction of {asset}")

        if asset in REQUIRING_URL_ASSETS:
            return list(self._fetch_and_add_url(asset))

        return list(self._query_all(queries[asset]))



from datetime import datetime

from .format import (
    DatabricksFormatter,
    _column_path,
    _column_payload,
    _table_payload,
    _to_datetime_or_none,
)


def test__to_datetime_or_none():
    time_ms = 1707734459947
    expected = datetime(2024, 2, 12, 10, 40, 59, 947000)

    assert _to_datetime_or_none(time_ms) == expected

    time_ms = None
    assert _to_datetime_or_none(time_ms) is None


def test_DatabricksFormatter__primary():
    emails = [
        {"type": "work", "value": "louis@evilcorp.com", "primary": False},
        {"type": "work", "value": "thomas@evilcorp.com", "primary": True},
    ]
    assert DatabricksFormatter._primary(emails) == "thomas@evilcorp.com"

    assert DatabricksFormatter._primary([]) is None


def test__table_payload():
    schema = {"id": "id123"}

    table = {
        "name": "baz",
        "catalog_name": "foo",
        "schema_name": "bar",
        "table_type": "MANAGED",
        "owner": "pot@ato.com",
        "table_id": "732pot5e-8ato-4c27-b701-9fa51febc192",
    }
    host = "https://some.cloud.databricks.net/"
    workspace_id = "123456"

    tags = {
        "foo.bar.baz": ["riri", "fifi"],
        "dummy.path": ["loulou"],
    }

    payload = _table_payload(schema, table, host, workspace_id, tags)

    expected = {
        "description": None,
        "id": "732pot5e-8ato-4c27-b701-9fa51febc192",
        "owner_email": "pot@ato.com",
        "schema_id": "id123",
        "table_name": "baz",
        "tags": ["riri", "fifi"],
        "type": "MANAGED",
        "url": "https://some.cloud.databricks.net/explore/data/foo/bar/baz?o=123456",
    }
    assert payload == expected


def test__column_payload():
    table = {
        "catalog_name": "foo",
        "name": "baz",
        "owner": "pot@ato.com",
        "schema_name": "bar",
        "table_id": "732pot5e-8ato-4c27-b701-9fa51febc192",
        "table_type": "MANAGED",
    }
    column = {
        "comment": "some description",
        "name": "Uid",
        "nullable": True,
        "position": 0,
        "type_json": '{"name":"Uid","type":"string","nullable":true,"metadata":{}}',
        "type_name": "STRING",
        "type_precision": 0,
        "type_scale": 0,
        "type_text": "string",
    }
    tags = {
        "foo.bar.baz.Uid": ["riri", "fifi"],
        "dummy.path": ["loulou"],
    }
    payload = _column_payload(table, column, tags)

    expected = {
        "column_name": "Uid",
        "data_type": "STRING",
        "description": "some description",
        "id": "`732pot5e-8ato-4c27-b701-9fa51febc192`.`Uid`",
        "ordinal_position": 0,
        "table_id": "732pot5e-8ato-4c27-b701-9fa51febc192",
        "tags": ["riri", "fifi"],
    }
    assert payload == expected

    # case where there are spaces in the name
    column["name"] = "column name with spaces"
    payload = _column_payload(table, column, tags)
    expected_id = (
        "`732pot5e-8ato-4c27-b701-9fa51febc192`.`column name with spaces`"
    )
    assert payload["id"] == expected_id


def test__column_path():
    table = {
        "catalog_name": "Jo",
        "schema_name": "William",
        "name": "Jack",
    }
    column = {
        "name": "Averell",
    }

    expected = "Jo.William.Jack.Averell"
    assert _column_path(table=table, column=column) == expected



"""
    File regrouping all constants used in PowerBi client
"""

DEFAULT_TIMEOUT_IN_SECS = 30
SCAN_READY = "Succeeded"
# ModifiedSince params should not be older than 30 days
RECENT_DAYS = 30

GET = "GET"
POST = "POST"


class Urls:
    """PowerBi's urls"""

    CLIENT_APP_BASE = "https://login.microsoftonline.com/"
    DEFAULT_SCOPE = "https://analysis.windows.net/powerbi/api/.default"
    REST_API_BASE_PATH = "https://api.powerbi.com/v1.0/myorg"

    # PBI rest API Routes
    ACTIVITY_EVENTS = f"{REST_API_BASE_PATH}/admin/activityevents"
    DASHBOARD = f"{REST_API_BASE_PATH}/admin/dashboards"
    DATASETS = f"{REST_API_BASE_PATH}/admin/datasets"
    GROUPS = f"{REST_API_BASE_PATH}/admin/groups"
    METADATA_GET = f"{REST_API_BASE_PATH}/admin/workspaces/scanResult"
    METADATA_POST = f"{REST_API_BASE_PATH}/admin/workspaces/getInfo"
    METADATA_WAIT = f"{REST_API_BASE_PATH}/admin/workspaces/scanStatus"
    REPORTS = f"{REST_API_BASE_PATH}/admin/reports"
    WORKSPACE_IDS = (
        "https://api.powerbi.com/v1.0/myorg/admin/workspaces/modified"
    )


class Batches:
    """Batches used within PowerBI api calls"""

    DEFAULT = 100
    # The route we use to fetch workspaces info can retrieve a maximum of
    # 100 workspaces per call
    # More: https://learn.microsoft.com/en-us/rest/api/power-bi/admin/workspace-info-post-workspace-info#request-body
    METADATA = 100


class QueryParams:
    """
    Frequently used PowerBi query params
    """

    METADATA_SCAN = {
        "datasetExpressions": True,
        "datasetSchema": True,
        "datasourceDetails": True,
        "getArtifactUsers": True,
        "lineage": True,
    }
    ACTIVE_WORKSPACE_FILTER = "state eq 'Active' and type eq 'Workspace'"


class Keys:
    ACCESS_TOKEN = "access_token"  # noqa: S105
    ACTIVITY_EVENT_ENTITIES = "activityEventEntities"
    CONTINUATION_URI = "continuationUri"
    ID = "id"
    INACTIVE_WORKSPACES = "excludeInActiveWorkspaces"
    LAST_RESULT_SET = "lastResultSet"
    MODIFIED_SINCE = "modifiedSince"
    PERSONAL_WORKSPACES = "excludePersonalWorkspaces"
    STATUS = "status"
    VALUE = "value"
    WORKSPACES = "workspaces"


class Assertions:
    """Assertion's messages"""

    BATCH_TOO_BIG = f"Can not retrieve more than {Batches.METADATA} at the time"
    DATETIME_TOO_OLD = "Date must be within 30 days range"



from typing import Any, DefaultDict, Dict, List, Sequence

from .object import getproperty
from .type import Getter


def group_by(identifier: Getter, elements: Sequence) -> Dict[Any, List]:
    """Groups the elements by the given key"""
    groups: Dict[Any, List] = DefaultDict(list)
    for element in elements:
        key = getproperty(element, identifier)
        groups[key].append(element)

    return groups


def mapping_from_rows(rows: List[Dict], key: Any, value: Any) -> Dict:
    """
    Create a dictionary mapping from a list of dictionaries using specified keys for mapping.

    Args:
        rows (list[dict]): A list of dictionaries from which to create the mapping.
        key (Any): The key to use for the keys of the resulting dictionary.
        value (Any): The key to use for the values of the resulting dictionary.

    Returns:
        dict: A dictionary where each key-value pair corresponds to the specified key and value
              from each dictionary in the input list. Only dictionaries with both specified key
              and value present are included in the result.

    Example:
        rows = [{'id': 1, 'name': 'Alice'}, {'id': 2, 'name': 'Bob'}]
        mapping = mapping_from_rows(rows, 'id', 'name')
        # mapping will be {1: 'Alice', 2: 'Bob'}
    """
    mapping = {}

    for row in rows:
        mapping_key = row.get(key)
        mapping_value = row.get(value)

        if not mapping_key or not mapping_value:
            continue
        mapping[mapping_key] = mapping_value

    return mapping



from enum import Enum

from ..utils import RetryStrategy

# url of the gcs proxy
INGEST_URL = "https://ingest.castordoc.com"

RETRY_BASE_MS = 10_000
RETRY_JITTER_MS = 1_000
RETRY_STRATEGY = RetryStrategy.LINEAR


class FileType(Enum):
    """type of file to load"""

    DBT = "DBT"
    QUALITY = "QUALITY"
    VIZ = "VIZ"
    WAREHOUSE = "WAREHOUSE"


PATH_TEMPLATES = {
    FileType.DBT: "transformation-{source_id}/{timestamp}-manifest.json",
    FileType.QUALITY: "quality-{source_id}/{timestamp}-{filename}",
    FileType.VIZ: "visualization-{source_id}/{filename}",
    FileType.WAREHOUSE: "warehouse-{source_id}/{filename}",
}


"""
The default request timeout in seconds for the upload
"""
DEFAULT_TIMEOUT = 60.0
ENVIRON_TIMEOUT = "CASTOR_TIMEOUT_OVERRIDE"

"""
The default retry for the upload
"""
DEFAULT_RETRY = 1
ENVIRON_RETRY = "CASTOR_RETRY_OVERRIDE"



DEFAULT_API_VERSION = 59.0
DEFAULT_PAGINATION_LIMIT = 100


class Keys:
    """Salesforce's credentials keys"""

    USERNAME = "username"
    PASSWORD = "password"  # noqa: S105
    CLIENT_ID = "client_id"
    CLIENT_SECRET = "client_secret"  # noqa: S105
    SECURITY_TOKEN = "security_token"  # noqa: S105
    BASE_URL = "base_url"



from enum import Enum
from typing import Dict

from requests.auth import HTTPBasicAuth

from ....utils import from_env


class CredentialsKey(Enum):
    HOST = "host"
    WORKSPACE = "workspace"
    TOKEN = "token"  # noqa: S105
    SECRET = "secret"  # noqa: S105


CREDENTIALS_ENV: Dict[CredentialsKey, str] = {
    CredentialsKey.HOST: "CASTOR_MODE_ANALYTICS_HOST",
    CredentialsKey.WORKSPACE: "CASTOR_MODE_ANALYTICS_WORKSPACE",
    CredentialsKey.TOKEN: "CASTOR_MODE_ANALYTICS_TOKEN",
    CredentialsKey.SECRET: "CASTOR_MODE_ANALYTICS_SECRET",
}


def get_value(key: CredentialsKey, kwargs: dict) -> str:
    """
    Returns the value of the given key:
    - from kwargs in priority
    - from ENV if not provided (raises an error if not found in ENV)
    """
    env_key = CREDENTIALS_ENV[key]
    return str(kwargs.get(key.value) or from_env(env_key))


class Credentials:
    """ValueObject for the credentials"""

    def __init__(self, host: str, workspace: str, token: str, secret: str):
        self.host = host
        self.workspace = workspace
        self.token = token
        self.secret = secret

    def to_dict(self, hide: bool = False) -> Dict[str, str]:
        safe = (
            CredentialsKey.HOST,
            CredentialsKey.WORKSPACE,
            CredentialsKey.TOKEN,
        )
        unsafe = (CredentialsKey.SECRET,)

        def val(k: CredentialsKey, v: str) -> str:
            return "*" + v[-4:] if hide and k in unsafe else v

        return {a.value: val(a, getattr(self, a.value)) for a in safe + unsafe}

    def authentication(self) -> HTTPBasicAuth:
        return HTTPBasicAuth(self.token, self.secret)



import logging
from typing import Any, Callable, Dict, Literal, Optional

import requests

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT_S = 30

# https://requests.readthedocs.io/en/latest/api/#requests.request
HttpMethod = Literal["GET", "OPTIONS", "HEAD", "POST", "PUT", "PATCH", "DELETE"]


class APIClient:
    """
    API client
    - authentication via access token
    """

    def __init__(self, host: str, token: Optional[str] = None):
        self._host = host
        self._token = token or ""
        self._timeout = DEFAULT_TIMEOUT_S

    @staticmethod
    def build_url(host: str, path: str):
        if not host.startswith("https://"):
            host = "https://" + host
        return f"{host.strip('/')}/{path}"

    def _headers(self) -> Dict[str, str]:
        if self._token:
            return {"Authorization": f"Bearer {self._token}"}
        return dict()

    def _call(
        self,
        url: str,
        method: HttpMethod = "GET",
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        processor: Optional[Callable] = None,
    ) -> Any:
        logger.debug(f"Calling {method} on {url}")
        result = requests.request(
            method,
            url,
            headers=self._headers(),
            params=params,
            json=data,
            timeout=self._timeout,
        )
        result.raise_for_status()

        if processor:
            return processor(result)

        return result.json()

    def get(
        self,
        path: str,
        payload: Optional[dict] = None,
        processor: Optional[Callable] = None,
    ) -> dict:
        """path: REST API operation path, such as /api/2.0/clusters/get"""
        url = self.build_url(self._host, path)
        return self._call(url=url, data=payload, processor=processor)



from .utils import build_path, tag_label


def test_build_path():
    row = {
        "bigflo": "oli",
        "laurel": "hardy",
        "dupond": "dupont",
    }
    keys = ["laurel", "dupond"]
    assert build_path(row, keys) == "hardy.dupont"


def test_tag_label():
    row = {
        "tag_name": "marketplace",
        "tag_value": "",
    }
    assert tag_label(row) == "marketplace"

    row = {
        "tag_name": "fi",
        "tag_value": "fou",
    }
    assert tag_label(row) == "fi:fou"



from ...types import ExternalAsset


class SigmaAsset(ExternalAsset):
    """Sigma assets"""

    DATASETS = "datasets"
    ELEMENTS = "elements"
    FILES = "files"
    LINEAGES = "lineages"
    MEMBERS = "members"
    QUERIES = "queries"
    WORKBOOKS = "workbooks"



import pytest

from .collection import mapping_from_rows


def test__mapping_from_rows__basic_mapping():
    rows = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice", 2: "Bob"}
    assert result == expected


def test__mapping_from_rows__missing_key():
    rows = [{"id": 1, "name": "Alice"}, {"name": "Bob"}]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice"}
    assert result == expected


def test__mapping_from_rows__missing_value():
    rows = [{"id": 1, "name": "Alice"}, {"id": 2}]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice"}
    assert result == expected


def test__mapping_from_rows__empty_list():
    rows = []
    result = mapping_from_rows(rows, "id", "name")
    expected = {}
    assert result == expected


def test__mapping_from_rows__non_existent_key_value():
    rows = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    result = mapping_from_rows(rows, "nonexistent_key", "nonexistent_value")
    expected = {}
    assert result == expected


def test__mapping_from_rows__none_key_value():
    rows = [
        {"id": 1, "name": "Alice"},
        {"id": None, "name": "Bob"},
        {"id": 2, "name": None},
    ]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice"}
    assert result == expected


def test__mapping_from_rows__multiple_valid_rows():
    rows = [
        {"id": 1, "name": "Alice"},
        {"id": 2, "name": "Bob"},
        {"id": 3, "name": "Charlie"},
    ]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice", 2: "Bob", 3: "Charlie"}
    assert result == expected
